#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Mock-up EoS module for testing generic property packages
"""
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.environ import Var, sqrt, units as pyunits

from idaes.models.properties.modular_properties.eos.eos_base import EoSBase


class DummyEoS(EoSBase):
    # Add attribute indicating support for electrolyte systems
    electrolyte_support = True

    @staticmethod
    def common(b, pobj):
        # Create dummy var to be returned by expression calls
        # This Var is used to create expressions where required.
        if not hasattr(b, "dummy_var"):
            b.dummy_var = Var(initialize=42)

        # Counter for how many times this method is called
        # This is used to ensure that the method has been called by checking
        # that the counter has advanced
        if hasattr(b, "eos_common"):
            b.eos_common += 1
        else:
            b.eos_common = 1

    @staticmethod
    def calculate_scaling_factors(b, pobj):
        pass

    @staticmethod
    def build_parameters(b):
        if not hasattr(b, "dummy_param"):
            b.dummy_param = Var(initialize=42)

    @staticmethod
    def act_phase_comp(b, p, j):
        return 42

    @staticmethod
    def act_coeff_phase_comp(b, p, j):
        return 1

    @staticmethod
    def compress_fact_phase(b, p):
        return 42

    @staticmethod
    def cp_mass_phase(b, p):
        return 42

    @staticmethod
    def cp_mol_phase(b, p):
        return 42

    @staticmethod
    def cp_mol_phase_comp(b, p, j):
        return 42

    @staticmethod
    def cv_mass_phase(b, p):
        return 42

    @staticmethod
    def cv_mol_phase(b, p):
        return 42

    @staticmethod
    def cv_mol_phase_comp(b, p, j):
        return 42

    @staticmethod
    def dens_mass_phase(b, p):
        return 42

    @staticmethod
    def dens_mol_phase(b, p):
        return 55e3 * pyunits.mol / pyunits.m**3

    @staticmethod
    def energy_internal_mol_phase(b, p):
        return 2e2 * b.temperature

    @staticmethod
    def energy_internal_mol_phase_comp(b, p, j):
        return 2e2 * b.temperature

    @staticmethod
    def enth_mol_phase(b, p):
        return 1e2 * pyunits.J / pyunits.mol / pyunits.K * b.temperature

    @staticmethod
    def enth_mol_phase_comp(b, p, j):
        return 1e2 * b.temperature

    @staticmethod
    def entr_mol_phase(b, p):
        return 42

    @staticmethod
    def entr_mol_phase_comp(b, p, j):
        return 42

    @staticmethod
    def fug_phase_comp(b, p, j):
        return 42

    @staticmethod
    def fug_coeff_phase_comp(b, p, j):
        return 42

    @staticmethod
    def gibbs_mol_phase(b, p):
        return 42

    @staticmethod
    def gibbs_mol_phase_comp(b, p, j):
        return 42

    @staticmethod
    def isothermal_speed_sound_phase(b, p):
        return 250

    @staticmethod
    def isentropic_speed_sound_phase(b, p):
        return sqrt(b.heat_capacity_ratio_phase[p]) * b.isothermal_speed_sound_phase[p]

    @staticmethod
    def vol_mol_phase(b, p):
        return 42

    @staticmethod
    def vol_mol_phase_comp(b, p, j):
        return 42

    @staticmethod
    def log_act_phase_comp(b, p, j):
        return 1

    @staticmethod
    def log_act_phase_solvents(b, p):
        return 1



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Membrane 1D model
"""
__author__ = "Maojian Wang"

# pylint: disable=unused-import
import pytest

from pyomo.environ import (
    check_optimal_termination,
    assert_optimal_termination,
    ConcreteModel,
    value,
)
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.solvers import get_solver
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
)
from idaes.core.util import DiagnosticsToolbox
from idaes.models_extra.power_generation.properties.natural_gas_PR import (
    get_prop,
    EosType,
)
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models_extra.co2_capture_and_utilization.unit_models import (
    Membrane1D,
    MembraneFlowPattern,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config_countercurrent():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = GenericParameterBlock(
        **get_prop(["CO2", "H2O", "N2"], ["Vap"], eos=EosType.IDEAL),
        doc="Key flue gas property parameters",
    )

    m.fs.unit = Membrane1D(
        finite_elements=3,
        dynamic=False,
        sweep_flow=True,
        flow_type=MembraneFlowPattern.COUNTERCURRENT,
        feed_side={"property_package": m.fs.properties},
        sweep_side={"property_package": m.fs.properties},
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 7
    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup


@pytest.mark.unit
def test_congif_cocurrent():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = GenericParameterBlock(
        **get_prop(["CO2", "H2O", "N2"], ["Vap"], eos=EosType.IDEAL),
        doc="Key flue gas property parameters",
    )

    m.fs.unit = Membrane1D(
        finite_elements=3,
        dynamic=False,
        sweep_flow=True,
        flow_type=MembraneFlowPattern.COCURRENT,
        feed_side={"property_package": m.fs.properties},
        sweep_side={"property_package": m.fs.properties},
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 7
    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup


class TestMembrane:
    @pytest.fixture(scope="class")
    def membrane(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = GenericParameterBlock(
            **get_prop(["CO2", "H2O", "N2"], ["Vap"], eos=EosType.IDEAL),
            doc="Key flue gas property parameters",
        )
        m.fs.unit = Membrane1D(
            finite_elements=3,
            dynamic=False,
            sweep_flow=True,
            flow_type=MembraneFlowPattern.COUNTERCURRENT,
            feed_side={"property_package": m.fs.properties},
            sweep_side={"property_package": m.fs.properties},
        )

        m.fs.unit.permeance[:, :, "CO2"].fix(1500)
        m.fs.unit.permeance[:, :, "H2O"].fix(1500 / 25)
        m.fs.unit.permeance[:, :, "N2"].fix(1500 / 25)
        m.fs.unit.area.fix(100)
        m.fs.unit.length.fix(10)

        m.fs.unit.feed_side_inlet.flow_mol[0].fix(100)
        m.fs.unit.feed_side_inlet.temperature[0].fix(365)
        m.fs.unit.feed_side_inlet.pressure[0].fix(120000)
        m.fs.unit.feed_side_inlet.mole_frac_comp[0, "N2"].fix(0.76)
        m.fs.unit.feed_side_inlet.mole_frac_comp[0, "CO2"].fix(0.13)
        m.fs.unit.feed_side_inlet.mole_frac_comp[0, "H2O"].fix(0.11)

        m.fs.unit.sweep_side_inlet.flow_mol[0].fix(0.01)
        m.fs.unit.sweep_side_inlet.temperature[0].fix(300)
        m.fs.unit.sweep_side_inlet.pressure[0].fix(51325)
        m.fs.unit.sweep_side_inlet.mole_frac_comp[0, "H2O"].fix(0.9986)
        m.fs.unit.sweep_side_inlet.mole_frac_comp[0, "CO2"].fix(0.0003)
        m.fs.unit.sweep_side_inlet.mole_frac_comp[0, "N2"].fix(0.0001)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, membrane):
        assert hasattr(membrane.fs.unit, "feed_side_inlet")
        assert len(membrane.fs.unit.feed_side_inlet.vars) == 4
        assert hasattr(membrane.fs.unit.feed_side_inlet, "flow_mol")
        assert hasattr(membrane.fs.unit.feed_side_inlet, "mole_frac_comp")
        assert hasattr(membrane.fs.unit.feed_side_inlet, "temperature")
        assert hasattr(membrane.fs.unit.feed_side_inlet, "pressure")

        assert hasattr(membrane.fs.unit, "sweep_side_inlet")
        assert len(membrane.fs.unit.sweep_side_inlet.vars) == 4
        assert hasattr(membrane.fs.unit.sweep_side_inlet, "flow_mol")
        assert hasattr(membrane.fs.unit.sweep_side_inlet, "mole_frac_comp")
        assert hasattr(membrane.fs.unit.sweep_side_inlet, "temperature")
        assert hasattr(membrane.fs.unit.sweep_side_inlet, "pressure")

        assert hasattr(membrane.fs.unit, "feed_side_outlet")
        assert len(membrane.fs.unit.feed_side_outlet.vars) == 4
        assert hasattr(membrane.fs.unit.feed_side_outlet, "flow_mol")
        assert hasattr(membrane.fs.unit.feed_side_outlet, "mole_frac_comp")
        assert hasattr(membrane.fs.unit.feed_side_outlet, "temperature")
        assert hasattr(membrane.fs.unit.feed_side_outlet, "pressure")

        assert hasattr(membrane.fs.unit, "sweep_side_outlet")
        assert len(membrane.fs.unit.sweep_side_outlet.vars) == 4
        assert hasattr(membrane.fs.unit.sweep_side_outlet, "flow_mol")
        assert hasattr(membrane.fs.unit.sweep_side_outlet, "mole_frac_comp")
        assert hasattr(membrane.fs.unit.sweep_side_outlet, "temperature")
        assert hasattr(membrane.fs.unit.sweep_side_outlet, "pressure")

        assert hasattr(membrane.fs.unit, "mscontactor")
        assert hasattr(membrane.fs.unit, "permeability_calculation")
        assert hasattr(membrane.fs.unit, "isothermal_constraint")

        assert number_variables(membrane) == 157
        assert number_total_constraints(membrane) == 89
        assert number_unused_variables(membrane) == 28

    @pytest.mark.component
    def test_structural_issues(self, membrane):
        dt = DiagnosticsToolbox(membrane)
        dt.assert_no_structural_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, membrane):
        initializer = BlockTriangularizationInitializer(constraint_tolerance=2e-5)
        initializer.initialize(membrane.fs.unit)
        results = solver.solve(membrane)
        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, membrane):
        dt = DiagnosticsToolbox(membrane)
        dt.assert_no_numerical_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_feed_solution(self, membrane):
        assert pytest.approx(99.99, abs=1e-2) == value(
            membrane.fs.unit.feed_side_outlet.flow_mol[0]
        )
        assert pytest.approx(0.1299, abs=1e-4) == value(
            membrane.fs.unit.feed_side_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.11, abs=1e-4) == value(
            membrane.fs.unit.feed_side_outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(0.76, abs=1e-4) == value(
            membrane.fs.unit.feed_side_outlet.mole_frac_comp[0, "N2"]
        )

        assert pytest.approx(365, abs=1e-2) == value(
            membrane.fs.unit.feed_side_outlet.temperature[0]
        )
        assert pytest.approx(120000, abs=1e-2) == value(
            membrane.fs.unit.feed_side_outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_sweep_side_solution(self, membrane):
        assert pytest.approx(0.01006, abs=1e-4) == value(
            membrane.fs.unit.sweep_side_outlet.flow_mol[0]
        )
        assert pytest.approx(0.0070, abs=1e-4) == value(
            membrane.fs.unit.sweep_side_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.99120, abs=1e-4) == value(
            membrane.fs.unit.sweep_side_outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(0.001710, abs=1e-4) == value(
            membrane.fs.unit.sweep_side_outlet.mole_frac_comp[0, "N2"]
        )

        assert pytest.approx(365, abs=1e-2) == value(
            membrane.fs.unit.sweep_side_outlet.temperature[0]
        )
        assert pytest.approx(51325.0, abs=1e-2) == value(
            membrane.fs.unit.sweep_side_outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_enthalpy_balance(self, membrane):

        assert (
            abs(
                value(
                    (
                        membrane.fs.unit.feed_side_inlet.flow_mol[0]
                        * membrane.fs.unit.mscontactor.feed_side_inlet_state[
                            0
                        ].enth_mol_phase["Vap"]
                        + membrane.fs.unit.sweep_side_inlet.flow_mol[0]
                        * membrane.fs.unit.mscontactor.sweep_side_inlet_state[
                            0
                        ].enth_mol_phase["Vap"]
                        - membrane.fs.unit.feed_side_outlet.flow_mol[0]
                        * membrane.fs.unit.mscontactor.feed_side[0, 3].enth_mol_phase[
                            "Vap"
                        ]
                        - membrane.fs.unit.sweep_side_outlet.flow_mol[0]
                        * membrane.fs.unit.mscontactor.sweep_side[0, 1].enth_mol_phase[
                            "Vap"
                        ]
                    )
                )
            )
            <= 1e-6
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_material_balance(self, membrane):

        assert (
            abs(
                value(
                    (
                        membrane.fs.unit.feed_side_inlet.flow_mol[0]
                        + membrane.fs.unit.sweep_side_inlet.flow_mol[0]
                        - membrane.fs.unit.feed_side_outlet.flow_mol[0]
                        - membrane.fs.unit.sweep_side_outlet.flow_mol[0]
                    )
                )
            )
            <= 1e-3
        )



# coding: utf-8
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

# stdlib
import importlib
import pathlib
import os
import re
import time

# package
import idaes
import pytest

good_modname = re.compile(r"[a-zA-Z][a-zA-Z0-9_]*")


SKIP, OK, BAD = "skipped", "success", "failed"


# NOTE (@lbianchi-lbl): this file should be rewritten as a proper pytest plugin


def print_path_status(path, status, msg=""):
    print(f"{status.upper():8s} {path} {msg}")


def importr(root: pathlib.Path, max_sec=10):
    """Import r_ecursively from the given root path."""
    base, failures, total = root.parent, {}, 0
    # iterate over flattened list of all paths ending a Python source file
    for path in root.rglob("*.py"):
        # check that all path components are valid module names
        # - this is to skip directories like '.ipynb_checkpoints'
        bad = False
        for name in path.parts[:-1]:
            if name == os.path.sep:
                continue
            if (
                name.startswith("Workshop")  # skip workshop tutorials (1)
                or name.startswith("Module_")  # skip workshop tutorials (2)
                or name == "tests"  # skip all tests
                or not good_modname.match(name)  # no _bla
            ):
                bad = True

            if bad:
                break
        # stop if bad directory component or Python filename is invalid
        bad = bad or not good_modname.match(path.parts[-1][:-3])
        if bad:
            print_path_status(path, SKIP)
            continue
        # module is valid, try importing it now
        module_path = path.relative_to(base).with_suffix("")
        module_name = ".".join(module_path.parts)
        try:
            start = time.time()
            importlib.import_module(module_name)
            sec = time.time() - start
            if sec > max_sec:
                print_path_status(path, BAD, msg="time={sec:3.1f}s")
                raise ImportError(f"Import took too long ({sec:.1f}s)")
            print_path_status(path, OK)
        except ImportError as e:
            failures[module_name] = str(e)
            print_path_status(path, BAD, msg="import error")
        total += 1
    return failures, total


@pytest.mark.unit
def test_import():
    root_dir = pathlib.Path(idaes.__file__).parent
    failures, total = importr(root_dir)
    n = len(failures)
    assert n == 0, f"{n:d} failures in {total:d} tests: {failures}"



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import itertools
import pytest

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentMap
import pyomo.environ as pyo

from pyomo.contrib.incidence_analysis import (
    IncidenceGraphInterface,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.util.subsystems import ParamSweeper

import idaes.core as idaes
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models_extra.gas_distribution.properties.natural_gas import (
    NaturalGasParameterBlock,
)
from idaes.models_extra.gas_distribution.unit_models.compressor import (
    IsothermalCompressor,
)
from idaes.core.solvers import get_solver

"""
Test for the simple compressor.
"""


@pytest.mark.component
class TestCompressorValues(unittest.TestCase):
    """
    This class tests whether the compressor correctly computes power
    requirement.
    """

    # Here we record predicted values for power requirement calculated
    # from each set of inputs.
    # These values were calculated with a pure-Pyomo simulation of the
    # compressor.
    # This dict maps pressure (bar), boost pressure (bar), and flow
    # rate (SCM/hr) to power requirement (kW).
    # Flow rate will later be converted to kmol/hr.
    expected_power = {
        (30.0, 2.0, 30.0e4): 560.0079273449446,
        (30.0, 2.0, 40.0e4): 746.6772364599237,
        (30.0, 2.0, 50.0e4): 933.3465455749101,
        (30.0, 2.0, 60.0e4): 1120.0158546898892,
        (30.0, 5.0, 30.0e4): 1350.2398277565735,
        (30.0, 5.0, 40.0e4): 1800.3197703420956,
        (30.0, 5.0, 50.0e4): 2250.399712927625,
        (30.0, 5.0, 60.0e4): 2700.479655513147,
        (30.0, 10.0, 30.0e4): 2555.623582229353,
        (30.0, 10.0, 40.0e4): 3407.498109639142,
        (30.0, 10.0, 50.0e4): 4259.372637048931,
        (30.0, 10.0, 60.0e4): 5111.247164458706,
        (40.0, 2.0, 30.0e4): 422.65881875472405,
        (40.0, 2.0, 40.0e4): 563.5450916729678,
        (40.0, 2.0, 50.0e4): 704.4313645912043,
        (40.0, 2.0, 60.0e4): 845.3176375094481,
        (40.0, 5.0, 30.0e4): 1027.7481241674832,
        (40.0, 5.0, 40.0e4): 1370.3308322233133,
        (40.0, 5.0, 50.0e4): 1712.9135402791435,
        (40.0, 5.0, 60.0e4): 2055.4962483349664,
        (40.0, 10.0, 30.0e4): 1968.8273762191675,
        (40.0, 10.0, 40.0e4): 2625.1031682922257,
        (40.0, 10.0, 50.0e4): 3281.378960365284,
        (40.0, 10.0, 60.0e4): 3937.654752438335,
        (50.0, 2.0, 30.0e4): 339.4199310524127,
        (50.0, 2.0, 40.0e4): 452.5599080698812,
        (50.0, 2.0, 50.0e4): 565.6998850873497,
        (50.0, 2.0, 60.0e4): 678.8398621048254,
        (50.0, 5.0, 30.0e4): 829.6932850445155,
        (50.0, 5.0, 40.0e4): 1106.2577133926825,
        (50.0, 5.0, 50.0e4): 1382.8221417408495,
        (50.0, 5.0, 60.0e4): 1659.386570089031,
        (50.0, 10.0, 30.0e4): 1601.739808280763,
        (50.0, 10.0, 40.0e4): 2135.653077707684,
        (50.0, 10.0, 50.0e4): 2669.566347134605,
        (50.0, 10.0, 60.0e4): 3203.479616561526,
        (60.0, 2.0, 30.0e4): 283.57512892647355,
        (60.0, 2.0, 40.0e4): 378.1001719019623,
        (60.0, 2.0, 50.0e4): 472.62521487745107,
        (60.0, 2.0, 60.0e4): 567.1502578529471,
        (60.0, 5.0, 30.0e4): 695.670690390034,
        (60.0, 5.0, 40.0e4): 927.5609205200453,
        (60.0, 5.0, 50.0e4): 1159.4511506500567,
        (60.0, 5.0, 60.0e4): 1391.341380780068,
        (60.0, 10.0, 30.0e4): 1350.2398277565735,
        (60.0, 10.0, 40.0e4): 1800.3197703420956,
        (60.0, 10.0, 50.0e4): 2250.399712927625,
        (60.0, 10.0, 60.0e4): 2700.479655513147,
    }

    def test_calculate_power(self):
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        compressor_config = {
            "property_package": m.fs.properties,
        }
        m.fs.compressor = IsothermalCompressor(**compressor_config)

        time = m.fs.time
        t0 = m.fs.time.first()

        # Preliminary constants and units
        # TODO: Should this nominal density constant live on the model?
        # Probably, even though it's not really necessary.
        density = 0.72 * pyo.units.kg / pyo.units.m**3
        kghr = pyo.units.kg / pyo.units.hr
        K = pyo.units.K
        bar = pyo.units.bar
        mw = m.fs.compressor.inlet_state[t0].mw
        scmhr = pyo.units.m**3 / pyo.units.hr
        kW = pyo.units.kW

        pressure_list = [30.0 * bar, 40.0 * bar, 50.0 * bar, 60.0 * bar]
        boost_pressure_list = [2.0 * bar, 5.0 * bar, 10.0 * bar]
        flow_list = [30.0e4 * scmhr, 40.0e4 * scmhr, 50.0e4 * scmhr, 60.0e4 * scmhr]
        n_scen = len(pressure_list) * len(boost_pressure_list) * len(flow_list)
        j = next(iter(m.fs.properties.component_list))
        inlet_values = {
            "temperature": [],
            "flow_mol": [],
            "pressure": [],
            "mole_frac_comp[%s]" % j: [],
        }
        input_values = {
            "boost_pressure[%s]" % t0: [],
        }
        target_values = {
            "power[%s]" % t0: [],
        }
        for p, dp, f in itertools.product(
            pressure_list,
            boost_pressure_list,
            flow_list,
        ):

            # Redundant unit conversion, for sanity.
            p_val = pyo.value(pyo.units.convert(p, bar))
            dp_val = pyo.value(pyo.units.convert(dp, bar))
            f_val = pyo.value(pyo.units.convert(f, scmhr))
            target_values["power[%s]" % t0].append(
                self.expected_power[p_val, dp_val, f_val] * kW
            )

            # IDAES generic property package requires kmol/hr.
            # I would prefer kg/hr...
            # SCM/hr -> kg/hr -> kmol/hr
            f = f * density / mw
            inlet_values["temperature"].append(293.15 * K)
            inlet_values["flow_mol"].append(f)
            inlet_values["pressure"].append(p)
            inlet_values["mole_frac_comp[%s]" % j].append(1.0)
            input_values["boost_pressure[%s]" % t0].append(dp)
        state = m.fs.compressor.inlet_state[t0]
        inlet_values = ComponentMap(
            (state.find_component(key), val) for key, val in inlet_values.items()
        )
        input_values = ComponentMap(
            (m.fs.compressor.find_component(key), val)
            for key, val in input_values.items()
        )
        input_values.update(inlet_values)
        target_values = ComponentMap(
            (m.fs.compressor.find_component(key), val)
            for key, val in target_values.items()
        )

        ipopt = get_solver("ipopt")
        param_sweeper = ParamSweeper(
            n_scen,
            input_values,
            output_values=target_values,
            to_fix=list(input_values.keys()),
        )
        with param_sweeper:
            self.assertEqual(degrees_of_freedom(m), 0)
            for inputs, target in param_sweeper:
                ipopt.solve(m, tee=False)

                # Sanity check that inputs have been properly set.
                for var, val in inputs.items():
                    val = pyo.value(pyo.units.convert(val, var.get_units()))
                    self.assertEqual(val, var.value)

                # Check that calculated power requirements are close to what
                # we expect. Values are different due to discrepancy in cv
                # between paper and IDAES.
                for var, val in target.items():
                    val = pyo.value(pyo.units.convert(val, var.get_units()))
                    # self.assertAlmostEqual(val, var.value, delta=1e2)
                    # self.assertAlmostEqual(val, var.value, reltol=0.1)
                    # Forget the relative tolerance argument for
                    # assertAlmostEqual, so using pytest.approx for now...
                    self.assertEqual(val, pytest.approx(var.value, rel=0.1))
                    # We use a pretty generous margin here because the
                    # property package's heat capacity cv is different
                    # from that used in the paper. (Paper does not use
                    # exactly (cp - R))


@pytest.mark.unit
class TestSimpleCompressor(unittest.TestCase):
    def test_compressor_methane(self):
        """
        Test constructing the compressor with a simple methane
        property package.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        compressor_config = {
            "property_package": m.fs.properties,
        }
        m.fs.compressor = IsothermalCompressor(**compressor_config)

        m.fs.compressor.inlet_state[0].temperature.fix(300.0 * pyo.units.K)
        m.fs.compressor.inlet_state[0].pressure.fix(20.0 * pyo.units.bar)
        m.fs.compressor.boost_pressure[0].fix(10.0 * pyo.units.bar)
        m.fs.compressor.inlet_state[0].flow_mol.fix(1000.0)
        # Fix mole_frac as defined_state is True at inlet
        j = next(iter(m.fs.properties.component_list))
        m.fs.compressor.inlet_state[0].mole_frac_comp[j].fix(1.0)

        # This test asserts:
        # (a) consistent units
        # (b) zero degrees of freedom
        # (c) a perfect matching between constraints and variables
        assert_units_consistent(m)

        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        N, M = igraph.incidence_matrix.shape
        self.assertEqual(N, M)
        matching = igraph.maximum_matching()
        self.assertEqual(len(matching), N)


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base classes for process costing
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# This plays with some private attributes - most are necessary
# pylint: disable=protected-access

from functools import partial

import pyomo.environ as pyo
from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr

from idaes.core import UnitModelBlockData
from idaes.core.base.process_base import declare_process_block_class, ProcessBlockData
from idaes.core.util.misc import add_object_reference, StrEnum
from idaes.core.util.exceptions import ConfigurationError

import idaes.logger as idaeslog


# Set up logger
_log = idaeslog.getLogger(__name__)


# TODO: Other functionality: methods for listing costed and uncosted units
def register_idaes_currency_units():
    """
    Define conversion rates for US Dollars based on CE Index.
    Source: https://www.toweringskills.com/financial-analysis/cost-indices/
    """
    if (
        "USD_CE500" in pyo.units._pint_registry
        and "USD_CE394" in pyo.units._pint_registry
    ):
        # Assume that standard units have already been registered
        # Log a message and end
        _log.debug(
            "Standard base currency units (USD_CE500, USD_CE394) "
            "already appear in Pyomo unit registry. Assuming repeated "
            "call of register_idaes_currency_units."
        )
    else:
        pyo.units.load_definitions_from_strings(
            [
                "USD_CE500 = [currency]",
                "USD_CE394 = 500/394.0 * USD_CE500",
                "USD_1990 = 500/357.6 * USD_CE500",
                "USD_1991 = 500/361.3 * USD_CE500",
                "USD_1992 = 500/358.2 * USD_CE500",
                "USD_1993 = 500/359.2 * USD_CE500",
                "USD_1994 = 500/368.1 * USD_CE500",
                "USD_1995 = 500/381.1 * USD_CE500",
                "USD_1996 = 500/381.7 * USD_CE500",
                "USD_1997 = 500/386.5 * USD_CE500",
                "USD_1998 = 500/389.5 * USD_CE500",
                "USD_1999 = 500/390.6 * USD_CE500",
                "USD_2000 = 500/394.1 * USD_CE500",
                "USD_2001 = 500/394.3 * USD_CE500",
                "USD_2002 = 500/395.6 * USD_CE500",
                "USD_2003 = 500/402.0 * USD_CE500",
                "USD_2004 = 500/444.2 * USD_CE500",
                "USD_2005 = 500/468.2 * USD_CE500",
                "USD_2006 = 500/499.6 * USD_CE500",
                "USD_2007 = 500/525.4 * USD_CE500",
                "USD_2008 = 500/575.4 * USD_CE500",
                "USD_2009 = 500/521.9 * USD_CE500",
                "USD_2010 = 500/550.8 * USD_CE500",
                "USD_2011 = 500/585.7 * USD_CE500",
                "USD_2012 = 500/584.6 * USD_CE500",
                "USD_2013 = 500/567.3 * USD_CE500",
                "USD_2014 = 500/576.1 * USD_CE500",
                "USD_2015 = 500/556.8 * USD_CE500",
                "USD_2016 = 500/541.7 * USD_CE500",
                "USD_2017 = 500/567.5 * USD_CE500",
                "USD_2018 = 500/603.1 * USD_CE500",
                "USD_2019 = 500/607.5 * USD_CE500",
                "USD_2020 = 500/596.2 * USD_CE500",
                "USD_2021 = 500/708.0 * USD_CE500",
                "USD_2022 = 500/816.0 * USD_CE500",
                "USD_2023 = 500/797.9 * USD_CE500",
            ]
        )


class DefaultCostingComponents(StrEnum):
    """
    Costing components Enum
    """

    capital = "capital_cost"
    fixed = "fixed_operating_cost"
    variable = "variable_operating_cost"


def assert_flowsheet_costing_block(val):
    """Domain validator for fowhseet costing block attributes

    Args:
        val : value to be checked

    Returns:
        ConfigurationError if val is not an instance of FlowsheetCostingBlock
    """
    if isinstance(val, FlowsheetCostingBlockData):
        return val
    else:
        _log.error(
            f"Flowsheet costing block argument {val} should "
            "be an instance of a FlowsheetCostingBlock"
        )
        raise ConfigurationError(
            f"Flowsheet costing block argument {val} should "
            "be an instance of a FlowsheetCostingBlock"
        )


@declare_process_block_class("FlowsheetCostingBlock")
class FlowsheetCostingBlockData(ProcessBlockData):
    """
    IDAES Flowsheet Costing Block class.

    This class is the base class used to create IDAES Flowsheet Costing Blocks,
    which are used to define calculations of capital and operating costs for a
    process.

    Developers should create a derived version of this class which should
    contain a library of methods for costing capital equipment as well as
    definitions of standard currency conversions and material and utility
    costs.
    """

    # Map costing methods to unit model classes
    unit_mapping = {}

    def build(self):
        """
        Base build method for FlowsheetCostingBlocks.

        This method sets up the basic attributes expected for registering
        costing of unit operations and flows and calls the build_global_params
        method from the associated costing package.
        """
        super().build()

        # Set up attributes for registering units and flows
        self._registered_unit_costing = []
        self.flow_types = pyo.Set()
        self.used_flows = pyo.Set()
        self._registered_flows = {}
        self.defined_flows = {}

        # Set the base year for all costs
        self.base_currency = None
        self.base_period = pyo.units.year

        # Register unit mapping
        self._costing_methods_map = {}
        self._build_costing_methods_map()

        # Build global params
        self.build_global_params()

        # Verify that costing package has set key attributes
        if self.base_currency is None:
            raise ValueError(
                f"{self.name} - costing package has not specified the base "
                "currency units to use for costing."
            )

        # Register pre-defined flow types
        for f, c in self.defined_flows.items():
            self.register_flow_type(f, c)

    def build_global_params(self):
        """
        This is where any global parameters, such as Lang factors or
        coefficients for costing methods that should be shared across the
        process, should be declared. Sub-Blocks may be used to help organize
        parameters if required.

        Derived class must overload this method.
        """
        raise NotImplementedError(
            "Derived class has not defined a build_global_params method."
        )

    def build_process_costs(self):
        """
        This is where process wide costing correlations should be declared.
        The following aggregate costs are available for use in calculating
        these process-wide costs:

        1. self.aggregate_capital_cost
        2. self.aggregate_fixed_operating_cost
        3. self.aggregate_variable_operating_cost
        4. self.aggregate_flow_costs (indexed by flow type)

        Derived class must overload this method.
        """
        raise NotImplementedError(
            "Derived class has not defined a build_process_costs method."
        )

    def initialize_build(self):
        """
        This is where custom initialization procedures can be implemented for
        flowsheet level costing components.

        Derived class must overload this method.
        """
        raise NotImplementedError(
            "Derived class has not defined an initialize_build method."
        )

    def cost_process(self):
        """
        This method constructs the process-level costing components based on
        the registered unit operations and flows.

        This first aggregates the costs from all the registered units and
        flows, and then calls the build_process_costs method from the
        associated costing package.
        """
        self.aggregate_costs()

        self.build_process_costs()

    def cost_flow(self, flow_expr, flow_type):
        """
        This method registers a given flow component (Var or expression) for
        costing. All flows are required to be bounded to be non-negative (i.e.
        a lower bound equal to or greater than 0).

        Args:
            flow_expr: Pyomo Var or expression that represents a material flow
                that should be included in the process costing. Units are
                expected to be on a per time basis.
            flow_type: string identifying the material this flow represents.
                This string must be registered with the FlowsheetCostingBlock
                as a known flow type.

        Raises:
            ValueError if flow_type is not recognized.
            TypeError if flow_type is an indexed Var.
        """
        # TODO : Is there any (efficient) way we can check for multiple costing
        # of the same flow? Given we allow expressions, I don't think this is
        # possible
        if flow_type not in self.flow_types:
            raise ValueError(
                f"{flow_type} is not a recognized flow type. Please check "
                "your spelling and that the flow type has been registered with"
                " the FlowsheetCostingBlock."
            )

        if type(flow_expr) in pyo.native_types:
            # this is a constant?!?
            pass
        elif flow_expr.is_indexed():
            raise TypeError(
                f"{flow_expr.name} is an indexed component. Flow costing only "
                "supports unindexed components."
            )
        elif not flow_expr.is_potentially_variable():
            # this is a constant?!?
            pass
        elif flow_expr.is_expression_type():
            # Get bounds from expression
            ebounds = compute_bounds_on_expr(flow_expr)
            if ebounds[0] is None or ebounds[0] < 0:
                _log.warning(
                    "flow_expr is an expression with a lower bound of less "
                    "than zero. Costing requires that all flows have a lower "
                    "bound equal to or greater than zero to avoid negative "
                    "costs."
                )
        else:
            # this should be a Var
            if flow_expr.lb is None or flow_expr.lb < 0:
                _log.warning(
                    f"{flow_expr.name} has a lower bound of less than zero. "
                    "Costing requires that all flows have a lower bound "
                    "equal to or greater than zero to avoid negative costs."
                )

        self._registered_flows[flow_type].append(flow_expr)

    def initialize(self):
        """
        This method attempts to initialize all the costing components
        registered with this FlowsheetCostingBlock. First, each registered
        UnitModelCostingBlock is initialized, followed by the aggregate
        costing variables and finally any process level costing components.

        Args:
            None
        """
        for u in self._registered_unit_costing:
            # Call initialize on all associated costing blocks
            # TODO: Think about ways to separate initialization of unit costing
            # TODO: from flowsheet costing
            u.initialize()

        # Initialize aggregate cost vars
        for c in DefaultCostingComponents:
            var = getattr(self, f"aggregate_{c}")
            cons = getattr(self, f"aggregate_{c}_constraint")
            calculate_variable_from_constraint(var, cons)

        # Initialize aggregate flows and costs
        for f in self.used_flows:
            fvar = getattr(self, f"aggregate_flow_{f}")
            fconst = getattr(self, f"aggregate_flow_{f}_constraint")
            calculate_variable_from_constraint(fvar, fconst)

            calculate_variable_from_constraint(
                self.aggregate_flow_costs[f],
                self.aggregate_flow_costs_constraint[f],
            )

        # Call costing package initialization
        try:
            # TODO: More general approach for initializing costing
            self.initialize_build()
        except AttributeError:
            # Assume the package has no initialize method
            pass

    def register_flow_type(self, flow_type, cost):
        """
        This method allows users to register new material and utility flows
        with the FlowsheetCostingBlock for use when costing flows.
        This method creates a new `Var` on the FlowsheetCostingBlock named
        f`{flow_type}_cost` whose value is fixed to `cost`.

        If a component named f`{flow_type}_cost` already exists on the
        FlowsheetCostingBlock, then an error is raised unless f`{flow_type}_cost`
        is `cost`. If f`{flow_type}_cost` is `cost`, no error is raised and
        the existing component f`{flow_type}_cost` is used to cost the flow.

        Args:
            flow_type: string name to represent flow type
            cost: a Pyomo expression with units representing the flow cost
        """
        self.flow_types.add(flow_type)

        name = f"{flow_type}_cost"
        current_component = self.component(name)
        if current_component is not None:
            if current_component is not cost:
                raise RuntimeError(
                    f"Component {name} already exists on {self} but is not {cost}."
                )
            # now self.{flow_type}_cost is cost, so just use it
        else:
            # Create a Var to hold the cost
            # Units will be different between flows, so have to use scalar Vars
            fvar = pyo.Var(
                units=pyo.units.get_units(cost),
                doc=f"Cost parameter for {flow_type} flow",
            )
            self.add_component(name, fvar)
            fvar.fix(cost)

        self._registered_flows[flow_type] = []

    def aggregate_costs(self):
        """
        This method aggregates costs from all the unit models and flows
        registered with this FlowsheetCostingBlock and creates aggregate
        variables for these on the FlowsheetCostingBlock that can be used for
        further process-wide costing calculations.

        The following costing variables are aggregated from all the registered
        UnitModelCostingBlocks (if they exist):

        * capital_cost,
        * fixed_operating_cost, and
        * variable_operating_cost

        Additionally, aggregate flow variables are created for all registered
        flow types along with aggregate costs associated with each of these.

        Args:
            None
        """
        c_units = self.base_currency
        t_units = self.base_period

        self.aggregate_capital_cost = pyo.Var(units=c_units)

        @self.Constraint(doc="Aggregation constraint for capital cost")
        def aggregate_capital_cost_constraint(blk):
            e = 0
            for u in self._registered_unit_costing:
                # Allow for units that might only have a subset of cost Vars
                if hasattr(u, "capital_cost"):
                    e += pyo.units.convert(u.capital_cost, to_units=c_units)

            return blk.aggregate_capital_cost == e

        # Aggregate unit operating costs
        self.aggregate_fixed_operating_cost = pyo.Var(units=c_units / t_units)

        @self.Constraint(doc="Aggregation constraint for fixed O&M costs")
        def aggregate_fixed_operating_cost_constraint(blk):
            e = 0
            for u in self._registered_unit_costing:
                # Allow for units that might only have a subset of cost Vars
                if hasattr(u, "fixed_operating_cost"):
                    e += pyo.units.convert(
                        u.fixed_operating_cost, to_units=c_units / t_units
                    )

            return blk.aggregate_fixed_operating_cost == e

        self.aggregate_variable_operating_cost = pyo.Var(units=c_units / t_units)

        @self.Constraint(doc="Aggregation constraint for variable O&M costs")
        def aggregate_variable_operating_cost_constraint(blk):
            e = 0
            for u in self._registered_unit_costing:
                # Allow for units that might only have a subset of cost Vars
                if hasattr(u, "variable_operating_cost"):
                    e += pyo.units.convert(
                        u.variable_operating_cost, to_units=c_units / t_units
                    )

            return blk.aggregate_variable_operating_cost == e

        # Aggregate flows
        # Units of flows will not be consistent, need separate Vars
        def agg_flow_rule(blk, f, funits):
            e = 0
            for flow in blk._registered_flows[f]:
                e += pyo.units.convert(flow, to_units=funits)
            agg_flow = getattr(blk, f"aggregate_flow_{f}")
            return agg_flow == e

        for f in self.flow_types:
            # We will use the first costed flow as representative of the whole
            if len(self._registered_flows[f]) > 0:
                self.used_flows.add(f)
                f1 = self._registered_flows[f][0]
                funits = pyo.units.get_units(f1)
                agg_var = pyo.Var(units=funits, doc=f"Aggregate flow for {f}")
                self.add_component(f"aggregate_flow_{f}", agg_var)

                agg_const = pyo.Constraint(
                    rule=partial(agg_flow_rule, f=f, funits=funits)
                )

                self.add_component(f"aggregate_flow_{f}_constraint", agg_const)

        # TODO : More complex cost functions
        self.aggregate_flow_costs = pyo.Var(self.used_flows, units=c_units / t_units)

        @self.Constraint(
            self.used_flows, doc="Aggregation constraint for material flow costs"
        )
        def aggregate_flow_costs_constraint(blk, ftype):
            agg_var = getattr(blk, f"aggregate_flow_{ftype}")
            cost_var = getattr(blk, f"{ftype}_cost")
            return blk.aggregate_flow_costs[ftype] == (
                pyo.units.convert(agg_var * cost_var, to_units=c_units / t_units)
            )

    def _build_costing_methods_map(self):
        """
        This method takes the mapping of unit model classes to costing methods
        from the costing package and constructs a mapping of the associated
        unit model data classes which is used to look up default costing
        methods.
        """
        for unit_class, meth in self.__class__.unit_mapping.items():
            # Need to get unit data class from unit class
            data_class = unit_class._ComponentDataClass
            self._costing_methods_map[data_class] = meth

    def _get_costing_method_for(self, unit_model):
        """
        This method used the _costing_methods_map mapping to try to identify
        the default costing method to use for a given unit model based on
        the data class of the unit model.
        """
        for base in unit_model.__class__.__mro__:
            if base in self._costing_methods_map:
                return self._costing_methods_map[base]
        raise RuntimeError(
            f"Could not identify default costing method for {unit_model.name}."
            " This implies the unit model's class and parent classes do not "
            "exist in the default mapping provided by the costing package. "
            "Please provide a specific costing method for this unit."
        )

    # -------------------------------------------------------------------------
    def report(self):
        # Need a clean reporting function
        # This will need input from the costing package
        pass

    def display_registered_unit_costing(self):
        # Need a method for cleanly showing registered unit models for costing
        pass

    def display_registered_flow_types(self):
        # Need a method for cleanly showing registered flow types
        pass

    def display_registered_flows(self):
        # Need a method for cleanly showing registered flow for costing
        pass


@declare_process_block_class("UnitModelCostingBlock")
class UnitModelCostingBlockData(ProcessBlockData):
    """
    Class for constructing costing blocks for unit models.

    At this stage, the only purpose of this class it to provide a distinct
    type for type-checking.
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "flowsheet_costing_block",
        ConfigValue(
            domain=assert_flowsheet_costing_block,
            doc="Reference to associated FlowsheetCostingBlock to use.",
        ),
    )
    CONFIG.declare(
        "costing_method",
        ConfigValue(
            doc=(
                "Costing method to use for unit (default from Costing Package "
                "used if not provided)."
            )
        ),
    )
    CONFIG.declare(
        "costing_method_arguments",
        ConfigValue(
            default={}, domain=dict, doc="Arguments to be passed to the costing method."
        ),
    )

    def build(self):
        super().build()

        # Alias flowsheet costing block reference
        fcb = self.config.flowsheet_costing_block

        # Get reference to unit model
        unit_model = self.parent_block()

        # Check that parent is an instance of a UnitModelBlockData
        if UnitModelBlockData not in unit_model.__class__.__mro__:
            raise TypeError(
                f"{self.name} - parent object ({unit_model.name}) is not an "
                f"instance of a UnitModelBlockData object. "
                "UnitModelCostingBlocks can only be added to UnitModelBlocks."
            )

        # Check to see if unit model already has costing
        for b in unit_model.component_objects(pyo.Block, descend_into=False):
            if b is not self and isinstance(b, UnitModelCostingBlock):
                # Block already has costing, clean up and raise exception
                raise RuntimeError(
                    f"Unit model {unit_model.name} already has a costing block"
                    f" registered: {b.name}. Each unit may only have a single "
                    "UnitModelCostingBlock associated with it."
                )
        # Add block to unit model initialization order
        unit_model._initialization_order.append(self)

        # Register unit model with this costing package
        fcb._registered_unit_costing.append(self)

        # Get costing method if not provided
        method = self.config.costing_method
        if method is None:
            method = fcb._get_costing_method_for(unit_model)

        # Assign object references for costing package and unit model
        add_object_reference(self, "costing_package", fcb)
        add_object_reference(self, "unit_model", unit_model)

        # Call unit costing method
        method(self, **self.config.costing_method_arguments)

        # Check that costs are Vars and have lower bound of 0
        cost_vars = DefaultCostingComponents
        for v in cost_vars:
            try:
                cvar = getattr(self, v)
                if not isinstance(cvar, pyo.Var):
                    raise TypeError(
                        f"{unit_model.name} {v} component must be a Var. "
                        "Please check the costing package you are using to "
                        "ensure that all costing components are declared as "
                        "variables."
                    )
                elif cvar.lb is None or cvar.lb < 0:
                    _log.warning(
                        f"{unit_model.name} {v} component has a lower bound "
                        "less than zero. Be aware that this may result in "
                        "negative costs during optimization."
                    )
            except AttributeError:
                pass

    def initialize(self, *args, **kwargs):
        """
        Placeholder method for initialization
        """
        # TODO: Implement an initialization method
        # TODO: Need to have a general purpose method (block triangularisation?)
        # TODO: Should also allow registering custom methods

        # Vars and Constraints
        for c in DefaultCostingComponents:
            if hasattr(self, c):
                var = getattr(self, c)
                cons = getattr(self, f"{c}_constraint")
                calculate_variable_from_constraint(var, cons)

    def del_costing(self):
        # Need to clean up references before deletion

        # Alias flowsheet costing block and unit model (parent)
        fcb = self.config.flowsheet_costing_block

        # Clean up references from fsb and unit_model
        self.unit_model._initialization_order.remove(self)
        fcb._registered_unit_costing.remove(self)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "Douglas Allan"

import pytest
import numpy as np

import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.models.unit_models.heat_exchanger import HeatExchangerFlowPattern
import idaes.models_extra.power_generation.unit_models.soc_submodels as soc
import idaes.models_extra.power_generation.unit_models.soc_submodels.testing as soc_testing

solver = get_solver("ipopt")


def build_tester(cell, nt, nz):
    soc_testing._build_test_utility(
        cell,
        comp_dict={
            pyo.Var: {
                "current_density": nz * nt,
                "potential": nt,
                "temperature_z": nz * nt,
                "length_z": 1,
                "length_y": 1,
                "total_current": nt,
            },
            pyo.Constraint: {
                "mean_temperature_eqn": nz * nt,
                "potential_eqn": nz * nt,
                "no_heat_flux_fuel_interconnect_eqn": nz * nt,
                "no_heat_flux_oxygen_interconnect_eqn": nz * nt,
                "total_current_eqn": nt,
            },
            pyo.Expression: {
                "dz": nz,
                "xface_area": nz,
                "cell_area": 1,
                "voltage_drop_contact": nz * nt,
                "voltage_drop_ohmic": nz * nt,
                "electrical_work": 1,
                "average_current_density": nt,
                "voltage_drop_interconnect": nz * nt,
            },
        },
    )


@pytest.fixture
def model():
    time_set = [0]
    zfaces = np.linspace(0, 1, 4).tolist()
    xfaces_electrode = np.linspace(0, 1, 6).tolist()
    xfaces_electrolyte = np.linspace(0, 1, 8).tolist()

    fuel_comps = ["H2", "H2O", "N2"]
    fuel_triple_phase_boundary_stoich_dict = {
        "H2": -0.5,
        "H2O": 0.5,
        "Vac": 0.5,
        "O^2-": -0.5,
        "e^-": 1.0,
    }
    oxygen_comps = ["O2", "N2"]
    oxygen_triple_phase_boundary_stoich_dict = {
        "O2": -0.25,
        "N2": 0,
        "Vac": -0.5,
        "O^2-": 0.5,
        "e^-": -1.0,
    }

    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=time_set, time_units=pyo.units.s)
    m.fs.cell = soc.SolidOxideCell(
        has_holdup=True,
        control_volume_zfaces=zfaces,
        control_volume_xfaces_fuel_electrode=xfaces_electrode,
        control_volume_xfaces_oxygen_electrode=xfaces_electrode,
        control_volume_xfaces_electrolyte=xfaces_electrolyte,
        fuel_component_list=fuel_comps,
        fuel_triple_phase_boundary_stoich_dict=fuel_triple_phase_boundary_stoich_dict,
        inert_fuel_species_triple_phase_boundary=["N2"],
        oxygen_component_list=oxygen_comps,
        oxygen_triple_phase_boundary_stoich_dict=oxygen_triple_phase_boundary_stoich_dict,
        inert_oxygen_species_triple_phase_boundary=["N2"],
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        include_contact_resistance=True,
    )
    return m


@pytest.fixture
def model_no_contact_resistance():
    time_set = [0]
    zfaces = np.linspace(0, 1, 6).tolist()
    xfaces_electrode = np.linspace(0, 1, 4).tolist()
    xfaces_electrolyte = np.linspace(0, 1, 12).tolist()

    fuel_comps = ["H2", "H2O"]
    fuel_triple_phase_boundary_stoich_dict = {
        "H2": -0.5,
        "H2O": 0.5,
        "Vac": 0.5,
        "O^2-": -0.5,
        "e^-": 1,
    }
    oxygen_comps = ["O2", "N2", "H2O"]
    oxygen_triple_phase_boundary_stoich_dict = {
        "O2": -0.25,
        "H2O": 0,
        "Vac": -0.5,
        "O^2-": 0.5,
        "e^-": -1,
    }

    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=time_set, time_units=pyo.units.s)
    m.fs.cell = soc.SolidOxideCell(
        has_holdup=False,
        control_volume_zfaces=zfaces,
        control_volume_xfaces_fuel_electrode=xfaces_electrode,
        control_volume_xfaces_oxygen_electrode=xfaces_electrode,
        control_volume_xfaces_electrolyte=xfaces_electrolyte,
        fuel_component_list=fuel_comps,
        fuel_triple_phase_boundary_stoich_dict=fuel_triple_phase_boundary_stoich_dict,
        oxygen_component_list=oxygen_comps,
        oxygen_triple_phase_boundary_stoich_dict=oxygen_triple_phase_boundary_stoich_dict,
        inert_oxygen_species_triple_phase_boundary=["N2", "H2O"],
        flow_pattern=HeatExchangerFlowPattern.cocurrent,
        include_contact_resistance=False,
    )
    return m


@pytest.mark.build
@pytest.mark.unit
def test_build(model):
    cell = model.fs.cell
    nt = len(model.fs.time)
    nz = len(cell.zfaces) - 1
    build_tester(cell, nt, nz)

    channels = [cell.fuel_channel, cell.oxygen_channel]

    for chan in channels:
        assert cell.temperature_z is chan.temperature_z.referent
        assert cell.length_y is chan.length_y.referent
        assert cell.length_z is chan.length_z.referent

    contact_resistors = [
        cell.contact_interconnect_fuel_flow_mesh,
        cell.contact_interconnect_oxygen_flow_mesh,
        cell.contact_flow_mesh_fuel_electrode,
        cell.contact_flow_mesh_oxygen_electrode,
    ]

    for unit in contact_resistors:
        assert cell.temperature_z is unit.temperature_z.referent
        assert cell.current_density is unit.current_density.referent
        assert cell.length_y is unit.length_y.referent
        assert cell.length_z is unit.length_z.referent

    assert (
        cell.fuel_channel.heat_flux_x0
        is cell.contact_interconnect_fuel_flow_mesh.heat_flux_x1.referent
    )
    assert (
        cell.fuel_channel.temperature_deviation_x0
        is cell.contact_interconnect_fuel_flow_mesh.temperature_deviation_x.referent
    )

    assert (
        cell.fuel_channel.heat_flux_x1
        is cell.contact_flow_mesh_fuel_electrode.heat_flux_x0.referent
    )
    assert (
        cell.fuel_channel.temperature_deviation_x1
        is cell.contact_flow_mesh_fuel_electrode.temperature_deviation_x.referent
    )

    assert (
        cell.oxygen_channel.heat_flux_x1
        is cell.contact_interconnect_oxygen_flow_mesh.heat_flux_x0.referent
    )
    assert (
        cell.oxygen_channel.temperature_deviation_x1
        is cell.contact_interconnect_oxygen_flow_mesh.temperature_deviation_x.referent
    )

    assert (
        cell.oxygen_channel.heat_flux_x0
        is cell.contact_flow_mesh_oxygen_electrode.heat_flux_x1.referent
    )
    assert (
        cell.oxygen_channel.temperature_deviation_x0
        is cell.contact_flow_mesh_oxygen_electrode.temperature_deviation_x.referent
    )

    electrodes = [cell.fuel_electrode, cell.oxygen_electrode]

    for trode in electrodes:
        assert cell.temperature_z is trode.temperature_z.referent
        assert cell.current_density is trode.current_density.referent
        assert cell.length_y is trode.length_y.referent
        assert cell.length_z is trode.length_z.referent

    assert (
        cell.fuel_channel.temperature_deviation_x1
        is cell.fuel_electrode.temperature_deviation_x0.referent
    )
    assert (
        cell.contact_flow_mesh_fuel_electrode.heat_flux_x1
        is cell.fuel_electrode.heat_flux_x0.referent
    )
    assert (
        cell.fuel_channel.conc_mol_comp
        is cell.fuel_electrode.conc_mol_comp_ref.referent
    )
    assert (
        cell.fuel_channel.conc_mol_comp_deviation_x1
        is cell.fuel_electrode.conc_mol_comp_deviation_x0.referent
    )
    assert (
        cell.fuel_channel.dconc_mol_compdt
        is cell.fuel_electrode.dconc_mol_comp_refdt.referent
    )
    assert (
        cell.fuel_channel.material_flux_x1
        is cell.fuel_electrode.material_flux_x0.referent
    )

    assert (
        cell.oxygen_channel.temperature_deviation_x0
        is cell.oxygen_electrode.temperature_deviation_x1.referent
    )
    assert (
        cell.contact_flow_mesh_oxygen_electrode.heat_flux_x0
        is cell.oxygen_electrode.heat_flux_x1.referent
    )
    assert (
        cell.oxygen_channel.conc_mol_comp
        is cell.oxygen_electrode.conc_mol_comp_ref.referent
    )
    assert (
        cell.oxygen_channel.conc_mol_comp_deviation_x0
        is cell.oxygen_electrode.conc_mol_comp_deviation_x1.referent
    )
    assert (
        cell.oxygen_channel.dconc_mol_compdt
        is cell.oxygen_electrode.dconc_mol_comp_refdt.referent
    )
    assert (
        cell.oxygen_channel.material_flux_x0
        is cell.oxygen_electrode.material_flux_x1.referent
    )

    tpb_list = [cell.fuel_triple_phase_boundary, cell.oxygen_triple_phase_boundary]

    for tpb in tpb_list:
        assert cell.temperature_z is tpb.temperature_z.referent
        assert cell.current_density is tpb.current_density.referent
        assert cell.length_y is tpb.length_y.referent
        assert cell.length_z is tpb.length_z.referent

    assert (
        cell.fuel_triple_phase_boundary.temperature_deviation_x.referent
        is cell.fuel_electrode.temperature_deviation_x1
    )
    assert (
        cell.fuel_triple_phase_boundary.heat_flux_x0.referent
        is cell.fuel_electrode.heat_flux_x1
    )
    assert (
        cell.fuel_triple_phase_boundary.conc_mol_comp_ref.referent
        is cell.fuel_channel.conc_mol_comp
    )
    assert (
        cell.fuel_triple_phase_boundary.conc_mol_comp_deviation_x.referent
        is cell.fuel_electrode.conc_mol_comp_deviation_x1
    )

    assert (
        cell.oxygen_triple_phase_boundary.temperature_deviation_x.referent
        is cell.oxygen_electrode.temperature_deviation_x0
    )
    assert (
        cell.oxygen_triple_phase_boundary.heat_flux_x1.referent
        is cell.oxygen_electrode.heat_flux_x0
    )
    assert (
        cell.oxygen_triple_phase_boundary.conc_mol_comp_ref.referent
        is cell.oxygen_channel.conc_mol_comp
    )
    assert (
        cell.oxygen_triple_phase_boundary.conc_mol_comp_deviation_x.referent
        is cell.oxygen_electrode.conc_mol_comp_deviation_x0
    )

    assert cell.temperature_z is cell.electrolyte.temperature_z.referent
    assert cell.current_density is cell.electrolyte.current_density.referent
    assert cell.length_y is cell.electrolyte.length_y.referent
    assert cell.length_z is cell.electrolyte.length_z.referent
    assert (
        cell.fuel_electrode.temperature_deviation_x1
        is cell.electrolyte.temperature_deviation_x0.referent
    )
    assert (
        cell.oxygen_electrode.temperature_deviation_x0
        is cell.electrolyte.temperature_deviation_x1.referent
    )
    assert (
        cell.fuel_triple_phase_boundary.heat_flux_x1
        is cell.electrolyte.heat_flux_x0.referent
    )
    assert (
        cell.oxygen_triple_phase_boundary.heat_flux_x0
        is cell.electrolyte.heat_flux_x1.referent
    )


@pytest.mark.component
def test_units(model):
    assert_units_consistent(model)


@pytest.mark.build
@pytest.mark.unit
def test_build_no_contact_resistance(model_no_contact_resistance):
    cell = model_no_contact_resistance.fs.cell
    nt = len(model_no_contact_resistance.fs.time)
    nz = len(cell.zfaces) - 1
    build_tester(cell, nt, nz)

    channels = [cell.fuel_channel, cell.oxygen_channel]

    for chan in channels:
        assert cell.temperature_z is chan.temperature_z.referent
        assert cell.length_y is chan.length_y.referent
        assert cell.length_z is chan.length_z.referent

    electrodes = [cell.fuel_electrode, cell.oxygen_electrode]

    for trode in electrodes:
        assert cell.temperature_z is trode.temperature_z.referent
        assert cell.current_density is trode.current_density.referent
        assert cell.length_y is trode.length_y.referent
        assert cell.length_z is trode.length_z.referent

    assert (
        cell.fuel_channel.temperature_deviation_x1
        is cell.fuel_electrode.temperature_deviation_x0.referent
    )
    assert cell.fuel_channel.heat_flux_x1 is cell.fuel_electrode.heat_flux_x0.referent
    assert (
        cell.fuel_channel.conc_mol_comp
        is cell.fuel_electrode.conc_mol_comp_ref.referent
    )
    assert (
        cell.fuel_channel.conc_mol_comp_deviation_x1
        is cell.fuel_electrode.conc_mol_comp_deviation_x0.referent
    )
    assert (
        cell.fuel_channel.dconc_mol_compdt
        is cell.fuel_electrode.dconc_mol_comp_refdt.referent
    )
    assert (
        cell.fuel_channel.material_flux_x1
        is cell.fuel_electrode.material_flux_x0.referent
    )

    assert (
        cell.oxygen_channel.temperature_deviation_x0
        is cell.oxygen_electrode.temperature_deviation_x1.referent
    )
    assert (
        cell.oxygen_channel.heat_flux_x0 is cell.oxygen_electrode.heat_flux_x1.referent
    )
    assert (
        cell.oxygen_channel.conc_mol_comp
        is cell.oxygen_electrode.conc_mol_comp_ref.referent
    )
    assert (
        cell.oxygen_channel.conc_mol_comp_deviation_x0
        is cell.oxygen_electrode.conc_mol_comp_deviation_x1.referent
    )
    assert (
        cell.oxygen_channel.dconc_mol_compdt
        is cell.oxygen_electrode.dconc_mol_comp_refdt.referent
    )
    assert (
        cell.oxygen_channel.material_flux_x0
        is cell.oxygen_electrode.material_flux_x1.referent
    )

    tpb_list = [cell.fuel_triple_phase_boundary, cell.oxygen_triple_phase_boundary]

    for tpb in tpb_list:
        assert cell.temperature_z is tpb.temperature_z.referent
        assert cell.current_density is tpb.current_density.referent
        assert cell.length_y is tpb.length_y.referent
        assert cell.length_z is tpb.length_z.referent

    assert (
        cell.fuel_triple_phase_boundary.temperature_deviation_x.referent
        is cell.fuel_electrode.temperature_deviation_x1
    )
    assert (
        cell.fuel_triple_phase_boundary.heat_flux_x0.referent
        is cell.fuel_electrode.heat_flux_x1
    )
    assert (
        cell.fuel_triple_phase_boundary.conc_mol_comp_ref.referent
        is cell.fuel_channel.conc_mol_comp
    )
    assert (
        cell.fuel_triple_phase_boundary.conc_mol_comp_deviation_x.referent
        is cell.fuel_electrode.conc_mol_comp_deviation_x1
    )

    assert (
        cell.oxygen_triple_phase_boundary.temperature_deviation_x.referent
        is cell.oxygen_electrode.temperature_deviation_x0
    )
    assert (
        cell.oxygen_triple_phase_boundary.heat_flux_x1.referent
        is cell.oxygen_electrode.heat_flux_x0
    )
    assert (
        cell.oxygen_triple_phase_boundary.conc_mol_comp_ref.referent
        is cell.oxygen_channel.conc_mol_comp
    )
    assert (
        cell.oxygen_triple_phase_boundary.conc_mol_comp_deviation_x.referent
        is cell.oxygen_electrode.conc_mol_comp_deviation_x0
    )

    assert cell.temperature_z is cell.electrolyte.temperature_z.referent
    assert cell.current_density is cell.electrolyte.current_density.referent
    assert cell.length_y is cell.electrolyte.length_y.referent
    assert cell.length_z is cell.electrolyte.length_z.referent
    assert (
        cell.fuel_electrode.temperature_deviation_x1
        is cell.electrolyte.temperature_deviation_x0.referent
    )
    assert (
        cell.oxygen_electrode.temperature_deviation_x0
        is cell.electrolyte.temperature_deviation_x1.referent
    )
    assert (
        cell.fuel_triple_phase_boundary.heat_flux_x1
        is cell.electrolyte.heat_flux_x0.referent
    )
    assert (
        cell.oxygen_triple_phase_boundary.heat_flux_x0
        is cell.electrolyte.heat_flux_x1.referent
    )


@pytest.mark.component
def test_units_no_contact_resistance(model_no_contact_resistance):
    assert_units_consistent(model_no_contact_resistance)


def build_tester_interconnect(cell, nt, nz):
    soc_testing._build_test_utility(
        cell,
        comp_dict={
            pyo.Var: {
                "current_density": nz * nt,
                "potential": nt,
                "temperature_z": nz * nt,
                "length_z": 1,
                "length_y": 1,
                "total_current": nt,
            },
            pyo.Constraint: {
                "mean_temperature_eqn": nz * nt,
                "potential_eqn": nz * nt,
                "total_current_eqn": nt,
            },
            pyo.Expression: {
                "cell_area": 1,
                "dz": nz,
                "xface_area": nz,
                "voltage_drop_contact": nz * nt,
                "voltage_drop_ohmic": nz * nt,
                "electrical_work": nt,
                "average_current_density": nt,
                "voltage_drop_interconnect": nz * nt,
            },
        },
    )


@pytest.fixture
def model_contact_resistance_and_interconnect():
    time_set = [0]
    zfaces = np.linspace(0, 1, 4).tolist()
    xfaces_electrode = np.linspace(0, 1, 6).tolist()
    xfaces_electrolyte = np.linspace(0, 1, 8).tolist()

    fuel_comps = ["H2", "H2O", "N2"]
    fuel_triple_phase_boundary_stoich_dict = {
        "H2": -0.5,
        "H2O": 0.5,
        "Vac": 0.5,
        "O^2-": -0.5,
        "e^-": 1.0,
    }
    oxygen_comps = ["O2", "N2"]
    oxygen_triple_phase_boundary_stoich_dict = {
        "O2": -0.25,
        "N2": 0,
        "Vac": -0.5,
        "O^2-": 0.5,
        "e^-": -1.0,
    }

    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(
        dynamic=False,
        time_set=time_set,
        time_units=pyo.units.s,
    )
    m.fs.cell = soc.SolidOxideCell(
        has_holdup=True,
        control_volume_zfaces=zfaces,
        control_volume_xfaces_fuel_electrode=xfaces_electrode,
        control_volume_xfaces_oxygen_electrode=xfaces_electrode,
        control_volume_xfaces_electrolyte=xfaces_electrolyte,
        fuel_component_list=fuel_comps,
        fuel_triple_phase_boundary_stoich_dict=fuel_triple_phase_boundary_stoich_dict,
        inert_fuel_species_triple_phase_boundary=["N2"],
        oxygen_component_list=oxygen_comps,
        oxygen_triple_phase_boundary_stoich_dict=oxygen_triple_phase_boundary_stoich_dict,
        inert_oxygen_species_triple_phase_boundary=["N2"],
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        include_contact_resistance=True,
        flux_through_interconnect=True,
        control_volume_xfaces_interconnect=[0, 0.5, 1],
    )
    return m


@pytest.mark.build
@pytest.mark.unit
def test_build_contact_resistance_and_interconnect(
    model_contact_resistance_and_interconnect,
):
    cell = model_contact_resistance_and_interconnect.fs.cell
    nt = len(model_contact_resistance_and_interconnect.fs.time)
    nz = len(cell.zfaces) - 1
    build_tester_interconnect(cell, nt, nz)

    channels = [cell.fuel_channel, cell.oxygen_channel]

    for chan in channels:
        assert cell.temperature_z is chan.temperature_z.referent
        assert cell.length_y is chan.length_y.referent
        assert cell.length_z is chan.length_z.referent

    contact_resistors = [
        cell.contact_interconnect_fuel_flow_mesh,
        cell.contact_interconnect_oxygen_flow_mesh,
        cell.contact_flow_mesh_fuel_electrode,
        cell.contact_flow_mesh_oxygen_electrode,
    ]

    for unit in contact_resistors:
        assert cell.temperature_z is unit.temperature_z.referent
        assert cell.current_density is unit.current_density.referent
        assert cell.length_y is unit.length_y.referent
        assert cell.length_z is unit.length_z.referent

    assert (
        cell.fuel_channel.heat_flux_x0
        is cell.contact_interconnect_fuel_flow_mesh.heat_flux_x1.referent
    )
    assert (
        cell.fuel_channel.temperature_deviation_x0
        is cell.contact_interconnect_fuel_flow_mesh.temperature_deviation_x.referent
    )

    assert (
        cell.fuel_channel.heat_flux_x1
        is cell.contact_flow_mesh_fuel_electrode.heat_flux_x0.referent
    )
    assert (
        cell.fuel_channel.temperature_deviation_x1
        is cell.contact_flow_mesh_fuel_electrode.temperature_deviation_x.referent
    )

    assert (
        cell.oxygen_channel.heat_flux_x1
        is cell.contact_interconnect_oxygen_flow_mesh.heat_flux_x0.referent
    )
    assert (
        cell.oxygen_channel.temperature_deviation_x1
        is cell.contact_interconnect_oxygen_flow_mesh.temperature_deviation_x.referent
    )

    assert (
        cell.oxygen_channel.heat_flux_x0
        is cell.contact_flow_mesh_oxygen_electrode.heat_flux_x1.referent
    )
    assert (
        cell.oxygen_channel.temperature_deviation_x0
        is cell.contact_flow_mesh_oxygen_electrode.temperature_deviation_x.referent
    )

    electrodes = [cell.fuel_electrode, cell.oxygen_electrode]

    for trode in electrodes:
        assert cell.temperature_z is trode.temperature_z.referent
        assert cell.current_density is trode.current_density.referent
        assert cell.length_y is trode.length_y.referent
        assert cell.length_z is trode.length_z.referent

    assert (
        cell.fuel_channel.temperature_deviation_x1
        is cell.fuel_electrode.temperature_deviation_x0.referent
    )
    assert (
        cell.contact_flow_mesh_fuel_electrode.heat_flux_x1
        is cell.fuel_electrode.heat_flux_x0.referent
    )
    assert (
        cell.fuel_channel.conc_mol_comp
        is cell.fuel_electrode.conc_mol_comp_ref.referent
    )
    assert (
        cell.fuel_channel.conc_mol_comp_deviation_x1
        is cell.fuel_electrode.conc_mol_comp_deviation_x0.referent
    )
    assert (
        cell.fuel_channel.dconc_mol_compdt
        is cell.fuel_electrode.dconc_mol_comp_refdt.referent
    )
    assert (
        cell.fuel_channel.material_flux_x1
        is cell.fuel_electrode.material_flux_x0.referent
    )

    assert (
        cell.oxygen_channel.temperature_deviation_x0
        is cell.oxygen_electrode.temperature_deviation_x1.referent
    )
    assert (
        cell.contact_flow_mesh_oxygen_electrode.heat_flux_x0
        is cell.oxygen_electrode.heat_flux_x1.referent
    )
    assert (
        cell.oxygen_channel.conc_mol_comp
        is cell.oxygen_electrode.conc_mol_comp_ref.referent
    )
    assert (
        cell.oxygen_channel.conc_mol_comp_deviation_x0
        is cell.oxygen_electrode.conc_mol_comp_deviation_x1.referent
    )
    assert (
        cell.oxygen_channel.dconc_mol_compdt
        is cell.oxygen_electrode.dconc_mol_comp_refdt.referent
    )
    assert (
        cell.oxygen_channel.material_flux_x0
        is cell.oxygen_electrode.material_flux_x1.referent
    )

    tpb_list = [cell.fuel_triple_phase_boundary, cell.oxygen_triple_phase_boundary]

    for tpb in tpb_list:
        assert cell.temperature_z is tpb.temperature_z.referent
        assert cell.current_density is tpb.current_density.referent
        assert cell.length_y is tpb.length_y.referent
        assert cell.length_z is tpb.length_z.referent

    assert (
        cell.fuel_triple_phase_boundary.temperature_deviation_x.referent
        is cell.fuel_electrode.temperature_deviation_x1
    )
    assert (
        cell.fuel_triple_phase_boundary.heat_flux_x0.referent
        is cell.fuel_electrode.heat_flux_x1
    )
    assert (
        cell.fuel_triple_phase_boundary.conc_mol_comp_ref.referent
        is cell.fuel_channel.conc_mol_comp
    )
    assert (
        cell.fuel_triple_phase_boundary.conc_mol_comp_deviation_x.referent
        is cell.fuel_electrode.conc_mol_comp_deviation_x1
    )

    assert (
        cell.oxygen_triple_phase_boundary.temperature_deviation_x.referent
        is cell.oxygen_electrode.temperature_deviation_x0
    )
    assert (
        cell.oxygen_triple_phase_boundary.heat_flux_x1.referent
        is cell.oxygen_electrode.heat_flux_x0
    )
    assert (
        cell.oxygen_triple_phase_boundary.conc_mol_comp_ref.referent
        is cell.oxygen_channel.conc_mol_comp
    )
    assert (
        cell.oxygen_triple_phase_boundary.conc_mol_comp_deviation_x.referent
        is cell.oxygen_electrode.conc_mol_comp_deviation_x0
    )

    assert cell.temperature_z is cell.electrolyte.temperature_z.referent
    assert cell.current_density is cell.electrolyte.current_density.referent
    assert cell.length_y is cell.electrolyte.length_y.referent
    assert cell.length_z is cell.electrolyte.length_z.referent
    assert (
        cell.fuel_electrode.temperature_deviation_x1
        is cell.electrolyte.temperature_deviation_x0.referent
    )
    assert (
        cell.oxygen_electrode.temperature_deviation_x0
        is cell.electrolyte.temperature_deviation_x1.referent
    )
    assert (
        cell.fuel_triple_phase_boundary.heat_flux_x1
        is cell.electrolyte.heat_flux_x0.referent
    )
    assert (
        cell.oxygen_triple_phase_boundary.heat_flux_x0
        is cell.electrolyte.heat_flux_x1.referent
    )

    assert cell.temperature_z is cell.interconnect.temperature_z.referent
    assert cell.current_density is cell.interconnect.current_density.referent
    assert cell.length_y is cell.interconnect.length_y.referent
    assert cell.length_z is cell.interconnect.length_z.referent

    assert (
        cell.oxygen_channel.temperature_deviation_x1
        is cell.interconnect.temperature_deviation_x0.referent
    )
    assert (
        cell.fuel_channel.temperature_deviation_x0
        is cell.interconnect.temperature_deviation_x1.referent
    )
    assert (
        cell.contact_interconnect_oxygen_flow_mesh.heat_flux_x1
        is cell.interconnect.heat_flux_x0.referent
    )
    assert (
        cell.contact_interconnect_fuel_flow_mesh.heat_flux_x0
        is cell.interconnect.heat_flux_x1.referent
    )


@pytest.mark.component
def test_units_contact_resistance_and_interconnect(
    model_contact_resistance_and_interconnect,
):
    assert_units_consistent(model_contact_resistance_and_interconnect)


@pytest.fixture
def model_no_contact_resistance_but_interconnect():
    time_set = [0]
    zfaces = np.linspace(0, 1, 6).tolist()
    xfaces_electrode = np.linspace(0, 1, 4).tolist()
    xfaces_electrolyte = np.linspace(0, 1, 12).tolist()
    xfaces_interconnect = np.linspace(0, 1, 16).tolist()

    fuel_comps = ["H2", "H2O"]
    fuel_triple_phase_boundary_stoich_dict = {
        "H2": -0.5,
        "H2O": 0.5,
        "Vac": 0.5,
        "O^2-": -0.5,
        "e^-": 1,
    }
    oxygen_comps = ["O2", "N2", "H2O"]
    oxygen_triple_phase_boundary_stoich_dict = {
        "O2": -0.25,
        "H2O": 0,
        "Vac": -0.5,
        "O^2-": 0.5,
        "e^-": -1,
    }

    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(
        dynamic=False,
        time_set=time_set,
        time_units=pyo.units.s,
    )
    m.fs.cell = soc.SolidOxideCell(
        has_holdup=False,
        control_volume_zfaces=zfaces,
        control_volume_xfaces_fuel_electrode=xfaces_electrode,
        control_volume_xfaces_oxygen_electrode=xfaces_electrode,
        control_volume_xfaces_electrolyte=xfaces_electrolyte,
        fuel_component_list=fuel_comps,
        fuel_triple_phase_boundary_stoich_dict=fuel_triple_phase_boundary_stoich_dict,
        # inert_fuel_species_triple_phase_boundary=[], Test default
        oxygen_component_list=oxygen_comps,
        oxygen_triple_phase_boundary_stoich_dict=oxygen_triple_phase_boundary_stoich_dict,
        inert_oxygen_species_triple_phase_boundary=["N2", "H2O"],
        flow_pattern=HeatExchangerFlowPattern.cocurrent,
        flux_through_interconnect=True,
        control_volume_xfaces_interconnect=xfaces_interconnect,
    )
    return m


@pytest.mark.build
@pytest.mark.unit
def test_build_no_contact_resistance_but_interconnect(
    model_no_contact_resistance_but_interconnect,
):
    cell = model_no_contact_resistance_but_interconnect.fs.cell
    nt = len(model_no_contact_resistance_but_interconnect.fs.time)
    nz = len(cell.zfaces) - 1
    build_tester_interconnect(cell, nt, nz)

    channels = [cell.fuel_channel, cell.oxygen_channel]

    for chan in channels:
        assert cell.temperature_z is chan.temperature_z.referent
        assert cell.length_y is chan.length_y.referent
        assert cell.length_z is chan.length_z.referent

    electrodes = [cell.fuel_electrode, cell.oxygen_electrode]

    for trode in electrodes:
        assert cell.temperature_z is trode.temperature_z.referent
        assert cell.current_density is trode.current_density.referent
        assert cell.length_y is trode.length_y.referent
        assert cell.length_z is trode.length_z.referent

    assert (
        cell.fuel_channel.temperature_deviation_x1
        is cell.fuel_electrode.temperature_deviation_x0.referent
    )
    assert cell.fuel_channel.heat_flux_x1 is cell.fuel_electrode.heat_flux_x0.referent
    assert (
        cell.fuel_channel.conc_mol_comp
        is cell.fuel_electrode.conc_mol_comp_ref.referent
    )
    assert (
        cell.fuel_channel.conc_mol_comp_deviation_x1
        is cell.fuel_electrode.conc_mol_comp_deviation_x0.referent
    )
    assert (
        cell.fuel_channel.dconc_mol_compdt
        is cell.fuel_electrode.dconc_mol_comp_refdt.referent
    )
    assert (
        cell.fuel_channel.material_flux_x1
        is cell.fuel_electrode.material_flux_x0.referent
    )

    assert (
        cell.oxygen_channel.temperature_deviation_x0
        is cell.oxygen_electrode.temperature_deviation_x1.referent
    )
    assert (
        cell.oxygen_channel.heat_flux_x0 is cell.oxygen_electrode.heat_flux_x1.referent
    )
    assert (
        cell.oxygen_channel.conc_mol_comp
        is cell.oxygen_electrode.conc_mol_comp_ref.referent
    )
    assert (
        cell.oxygen_channel.conc_mol_comp_deviation_x0
        is cell.oxygen_electrode.conc_mol_comp_deviation_x1.referent
    )
    assert (
        cell.oxygen_channel.dconc_mol_compdt
        is cell.oxygen_electrode.dconc_mol_comp_refdt.referent
    )
    assert (
        cell.oxygen_channel.material_flux_x0
        is cell.oxygen_electrode.material_flux_x1.referent
    )

    tpb_list = [cell.fuel_triple_phase_boundary, cell.oxygen_triple_phase_boundary]

    for tpb in tpb_list:
        assert cell.temperature_z is tpb.temperature_z.referent
        assert cell.current_density is tpb.current_density.referent
        assert cell.length_y is tpb.length_y.referent
        assert cell.length_z is tpb.length_z.referent

    assert (
        cell.fuel_triple_phase_boundary.temperature_deviation_x.referent
        is cell.fuel_electrode.temperature_deviation_x1
    )
    assert (
        cell.fuel_triple_phase_boundary.heat_flux_x0.referent
        is cell.fuel_electrode.heat_flux_x1
    )
    assert (
        cell.fuel_triple_phase_boundary.conc_mol_comp_ref.referent
        is cell.fuel_channel.conc_mol_comp
    )
    assert (
        cell.fuel_triple_phase_boundary.conc_mol_comp_deviation_x.referent
        is cell.fuel_electrode.conc_mol_comp_deviation_x1
    )

    assert (
        cell.oxygen_triple_phase_boundary.temperature_deviation_x.referent
        is cell.oxygen_electrode.temperature_deviation_x0
    )
    assert (
        cell.oxygen_triple_phase_boundary.heat_flux_x1.referent
        is cell.oxygen_electrode.heat_flux_x0
    )
    assert (
        cell.oxygen_triple_phase_boundary.conc_mol_comp_ref.referent
        is cell.oxygen_channel.conc_mol_comp
    )
    assert (
        cell.oxygen_triple_phase_boundary.conc_mol_comp_deviation_x.referent
        is cell.oxygen_electrode.conc_mol_comp_deviation_x0
    )

    assert cell.temperature_z is cell.electrolyte.temperature_z.referent
    assert cell.current_density is cell.electrolyte.current_density.referent
    assert cell.length_y is cell.electrolyte.length_y.referent
    assert cell.length_z is cell.electrolyte.length_z.referent
    assert (
        cell.fuel_electrode.temperature_deviation_x1
        is cell.electrolyte.temperature_deviation_x0.referent
    )
    assert (
        cell.oxygen_electrode.temperature_deviation_x0
        is cell.electrolyte.temperature_deviation_x1.referent
    )
    assert (
        cell.fuel_triple_phase_boundary.heat_flux_x1
        is cell.electrolyte.heat_flux_x0.referent
    )
    assert (
        cell.oxygen_triple_phase_boundary.heat_flux_x0
        is cell.electrolyte.heat_flux_x1.referent
    )

    assert (
        cell.oxygen_channel.temperature_deviation_x1
        is cell.interconnect.temperature_deviation_x0.referent
    )
    assert (
        cell.fuel_channel.temperature_deviation_x0
        is cell.interconnect.temperature_deviation_x1.referent
    )
    assert cell.oxygen_channel.heat_flux_x1 is cell.interconnect.heat_flux_x0.referent
    assert cell.fuel_channel.heat_flux_x0 is cell.interconnect.heat_flux_x1.referent


@pytest.mark.component
def test_units_no_contact_resistance_but_interconnect(
    model_no_contact_resistance_but_interconnect,
):
    assert_units_consistent(model_no_contact_resistance_but_interconnect)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for scaling utility functions.

Author: Andrew Lee
"""
from io import StringIO
import os
import pytest
import re

from pyomo.environ import Block, Constraint, ConcreteModel, Set, Suffix, Var
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

from idaes.core.scaling.util import (
    get_scaling_suffix,
    get_scaling_factor,
    set_scaling_factor,
    del_scaling_factor,
    _suffix_to_dict,
    _suffix_from_dict,
    _collect_block_suffixes,
    _set_block_suffixes_from_dict,
    scaling_factors_to_dict,
    scaling_factors_from_dict,
    scaling_factors_to_json_file,
    scaling_factors_from_json_file,
    report_scaling_factors,
)
import idaes.logger as idaeslog

currdir = this_file_dir()


class TestGetScalingSuffix:
    @pytest.mark.unit
    def test_get_scaling_suffix_block_new(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        sfx = get_scaling_suffix(m)

        assert "Created new scaling suffix for unknown" in caplog.text

        assert isinstance(m.scaling_factor, Suffix)
        assert sfx is m.scaling_factor

    @pytest.mark.unit
    def test_get_scaling_suffix_indexed_component_new(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v = Var([1, 2, 3, 4])
        sfx = get_scaling_suffix(m.v[1])

        assert "Created new scaling suffix for unknown" in caplog.text

        assert isinstance(m.scaling_factor, Suffix)
        assert sfx is m.scaling_factor

    @pytest.mark.unit
    def test_get_scaling_suffix_indexed_block(self):
        m = ConcreteModel()
        m.b = Block([1, 2, 3, 4])

        with pytest.raises(
            TypeError,
            match="IndexedBlocks cannot have scaling factors attached to them. "
            "Please assign scaling factors to the elements of the IndexedBlock.",
        ):
            get_scaling_suffix(m.b)

    @pytest.mark.unit
    def test_get_scaling_suffix_component_new(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v = Var()
        sfx = get_scaling_suffix(m.v)

        assert "Created new scaling suffix for unknown" in caplog.text

        assert isinstance(m.scaling_factor, Suffix)
        assert sfx is m.scaling_factor

    @pytest.mark.unit
    def test_get_scaling_suffix_block_existing(self, caplog):
        m = ConcreteModel()
        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        sfx = get_scaling_suffix(m)

        assert "Created new scaling suffix for unknown" not in caplog.text

        assert isinstance(m.scaling_factor, Suffix)
        assert sfx is m.scaling_factor

    @pytest.mark.unit
    def test_get_scaling_suffix_component_existing(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        m.v = Var()
        sfx = get_scaling_suffix(m.v)

        assert "Created new scaling suffix for unknown" not in caplog.text

        assert isinstance(m.scaling_factor, Suffix)
        assert sfx is m.scaling_factor


class TestSuffixToFromDict:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.s = Set(initialize=[1, 2, 3, 4])

        m.v = Var(m.s)

        @m.Constraint(m.s)
        def c(b, i):
            return b.v[i] == i

        m.b = Block(m.s)

        for bd in m.b.values():
            bd.v2 = Var()

            bd.scaling_factor = Suffix(direction=Suffix.EXPORT)
            bd.scaling_factor[bd.v2] = 10

        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        for i in m.s:
            m.scaling_factor[m.v[i]] = 5 * i
            m.scaling_factor[m.c[i]] = 5**i

        return m

    @pytest.mark.unit
    def test_suffix_to_dict(self, model):
        sdict = _suffix_to_dict(model.scaling_factor)

        assert sdict == {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
        }

    @pytest.mark.unit
    def test_suffix_from_dict(self, model):
        sdict = {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
        }

        _suffix_from_dict(model.scaling_factor, sdict, overwrite=True)

        assert model.scaling_factor[model.v[1]] == 5
        assert model.scaling_factor[model.v[2]] == 10
        assert model.scaling_factor[model.v[3]] == 15
        assert model.scaling_factor[model.v[4]] == 20

        assert model.scaling_factor[model.c[1]] == 5
        assert model.scaling_factor[model.c[2]] == 25
        assert model.scaling_factor[model.c[3]] == 125
        assert model.scaling_factor[model.c[4]] == 625

        assert len(model.scaling_factor) == 8

    @pytest.mark.unit
    def test_suffix_from_dict_invalid_component_name(self, model):
        sdict = {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "foo": 7,
        }

        with pytest.raises(
            ValueError,
            match=re.escape("Could not find component foo on block unknown."),
        ):
            _suffix_from_dict(model.scaling_factor, sdict, overwrite=True)

        # If we set verify_name=False, it should proceed
        _suffix_from_dict(
            model.scaling_factor, sdict, overwrite=True, verify_names=False
        )

        assert model.scaling_factor[model.v[1]] == 5
        assert model.scaling_factor[model.v[2]] == 10
        assert model.scaling_factor[model.v[3]] == 15
        assert model.scaling_factor[model.v[4]] == 20

        assert model.scaling_factor[model.c[1]] == 5
        assert model.scaling_factor[model.c[2]] == 25
        assert model.scaling_factor[model.c[3]] == 125
        assert model.scaling_factor[model.c[4]] == 625

        assert len(model.scaling_factor) == 8

    @pytest.mark.unit
    def test_collect_block_suffixes_single(self):
        m = ConcreteModel()
        m.s = Set(initialize=[1, 2, 3, 4])

        m.v = Var(m.s)

        @m.Constraint(m.s)
        def c(b, i):
            return b.v[i] == i

        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        for i in m.s:
            m.scaling_factor[m.v[i]] = 5 * i
            m.scaling_factor[m.c[i]] = 5**i

        sdict = _collect_block_suffixes(m)

        assert sdict == {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "subblock_suffixes": {},
        }

    @pytest.mark.unit
    def test_collect_block_suffixes_nested(self, model):
        sdict = _collect_block_suffixes(model)

        assert sdict == {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "subblock_suffixes": {
                "b[1]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[2]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[3]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[4]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
            },
        }

    @pytest.mark.unit
    def test_collect_block_suffixes_nested_descend_false(self, model):
        sdict = _collect_block_suffixes(model, descend_into=False)

        assert sdict == {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
        }

    @pytest.mark.unit
    def test_set_block_suffixes_from_dict(self):
        m = ConcreteModel()
        m.s = Set(initialize=[1, 2, 3, 4])

        m.v = Var(m.s)

        @m.Constraint(m.s)
        def c(b, i):
            return b.v[i] == i

        m.b = Block(m.s)

        for bd in m.b.values():
            bd.v2 = Var()

        # Set suffix values to retrieve
        # Only set values for some subblocks to make sure behaviour is correct
        sdict = {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "subblock_suffixes": {
                "b[1]": {
                    "v2": 10,
                },
                "b[2]": {
                    "v2": 20,
                },
            },
        }

        _set_block_suffixes_from_dict(m, sdict)

        assert m.scaling_factor[m.v[1]] == 5
        assert m.scaling_factor[m.v[2]] == 10
        assert m.scaling_factor[m.v[3]] == 15
        assert m.scaling_factor[m.v[4]] == 20

        assert m.scaling_factor[m.c[1]] == 5
        assert m.scaling_factor[m.c[2]] == 25
        assert m.scaling_factor[m.c[3]] == 125
        assert m.scaling_factor[m.c[4]] == 625

        assert len(m.scaling_factor) == 8

        assert m.b[1].scaling_factor[m.b[1].v2] == 10
        assert len(m.b[1].scaling_factor) == 1

        assert m.b[2].scaling_factor[m.b[2].v2] == 20
        assert len(m.b[2].scaling_factor) == 1

        assert not hasattr(m.b[3], "scaling_factor")
        assert not hasattr(m.b[4], "scaling_factor")

    @pytest.mark.unit
    def test_set_block_suffixes_from_dict_overwrite_false(self):
        m = ConcreteModel()
        m.s = Set(initialize=[1, 2, 3, 4])

        m.v = Var(m.s)

        @m.Constraint(m.s)
        def c(b, i):
            return b.v[i] == i

        m.b = Block(m.s)

        for bd in m.b.values():
            bd.v2 = Var()

            # Set some existing scaling factors
            bd.scaling_factor = Suffix(direction=Suffix.EXPORT)
            bd.scaling_factor[bd.v2] = 100

        # Set suffix values to retrieve
        # Only set values for some subblocks to make sure behaviour is correct
        sdict = {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "subblock_suffixes": {
                "b[1]": {
                    "v2": 10,
                },
                "b[2]": {
                    "v2": 20,
                },
            },
        }

        _set_block_suffixes_from_dict(m, sdict, overwrite=False)

        assert m.scaling_factor[m.v[1]] == 5
        assert m.scaling_factor[m.v[2]] == 10
        assert m.scaling_factor[m.v[3]] == 15
        assert m.scaling_factor[m.v[4]] == 20

        assert m.scaling_factor[m.c[1]] == 5
        assert m.scaling_factor[m.c[2]] == 25
        assert m.scaling_factor[m.c[3]] == 125
        assert m.scaling_factor[m.c[4]] == 625

        assert len(m.scaling_factor) == 8

        for i in [1, 2, 3, 4]:
            assert m.b[i].scaling_factor[m.b[i].v2] == 100
            assert len(m.b[i].scaling_factor) == 1

        # Check that we did not mutate the original dict
        assert sdict == {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "subblock_suffixes": {
                "b[1]": {
                    "v2": 10,
                },
                "b[2]": {
                    "v2": 20,
                },
            },
        }

    @pytest.mark.unit
    def test_set_block_suffixes_from_dict_verify_names(self):
        m = ConcreteModel()
        m.s = Set(initialize=[1, 2, 3, 4])

        m.v = Var(m.s)

        @m.Constraint(m.s)
        def c(b, i):
            return b.v[i] == i

        m.b = Block(m.s)

        for bd in m.b.values():
            bd.v2 = Var()

        # Set suffix values to retrieve
        # Only set values for some subblocks to make sure behaviour is correct
        sdict = {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "subblock_suffixes": {
                "b[1]": {
                    "v2": 10,
                },
                "foo": {
                    "v2": 20,
                },
            },
        }

        with pytest.raises(
            AttributeError,
            match="Block unknown does not have a subblock named foo.",
        ):
            _set_block_suffixes_from_dict(m, sdict, verify_names=True)

    @pytest.mark.unit
    def test_scaling_factors_to_dict_suffix(self, model):
        sdict = scaling_factors_to_dict(model.scaling_factor)

        assert sdict == {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "block_name": "unknown",
        }

    @pytest.mark.unit
    def test_scaling_factors_to_dict_blockdata_descend_false(self, model):
        sdict = scaling_factors_to_dict(model, descend_into=False)

        assert sdict == {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "block_name": "unknown",
        }

    @pytest.mark.unit
    def test_scaling_factors_to_dict_blockdata_descend_true(self, model):
        sdict = scaling_factors_to_dict(model, descend_into=True)

        assert sdict == {
            "v[1]": 5,
            "v[2]": 10,
            "v[3]": 15,
            "v[4]": 20,
            "c[1]": 5,
            "c[2]": 25,
            "c[3]": 125,
            "c[4]": 625,
            "subblock_suffixes": {
                "b[1]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[2]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[3]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[4]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
            },
            "block_name": "unknown",
        }

    @pytest.mark.unit
    def test_scaling_factors_to_dict_indexed_block(self, model):
        sdict = scaling_factors_to_dict(model.b, descend_into=True)

        assert sdict == {
            "block_datas": {
                "b[1]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[2]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[3]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
                "b[4]": {
                    "v2": 10,
                    "subblock_suffixes": {},
                },
            },
            "block_name": "b",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_suffix(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "unknown",
        }

        scaling_factors_from_dict(
            model.scaling_factor, sdict, overwrite=True, verify_names=True
        )

        assert model.scaling_factor[model.v[1]] == 50
        assert model.scaling_factor[model.v[2]] == 100
        assert model.scaling_factor[model.v[3]] == 15
        assert model.scaling_factor[model.v[4]] == 20
        assert model.scaling_factor[model.c[1]] == 50
        assert model.scaling_factor[model.c[2]] == 250
        assert model.scaling_factor[model.c[3]] == 125
        assert model.scaling_factor[model.c[4]] == 625
        assert len(model.scaling_factor) == 8

        for bd in model.b.values():
            assert bd.scaling_factor[bd.v2] == 10
            assert len(bd.scaling_factor) == 1

        # Ensure we have not mutated original dict
        assert sdict == {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "unknown",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_suffix_overwrite_false(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "unknown",
        }

        scaling_factors_from_dict(
            model.scaling_factor, sdict, overwrite=False, verify_names=True
        )

        assert model.scaling_factor[model.v[1]] == 5
        assert model.scaling_factor[model.v[2]] == 10
        assert model.scaling_factor[model.v[3]] == 15
        assert model.scaling_factor[model.v[4]] == 20
        assert model.scaling_factor[model.c[1]] == 5
        assert model.scaling_factor[model.c[2]] == 25
        assert model.scaling_factor[model.c[3]] == 125
        assert model.scaling_factor[model.c[4]] == 625
        assert len(model.scaling_factor) == 8

        for bd in model.b.values():
            assert bd.scaling_factor[bd.v2] == 10
            assert len(bd.scaling_factor) == 1

        # Ensure we have not mutated original dict
        assert sdict == {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "unknown",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_suffix_verify_fail(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "foo",
        }

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Name of parent block (unknown) does not match that recorded in json_dict (foo)"
            ),
        ):
            scaling_factors_from_dict(
                model.scaling_factor, sdict, overwrite=True, verify_names=True
            )

        # Ensure we have not mutated original dict
        assert sdict == {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "foo",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_block_data(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "unknown",
        }

        scaling_factors_from_dict(model, sdict, overwrite=True, verify_names=True)

        assert model.scaling_factor[model.v[1]] == 50
        assert model.scaling_factor[model.v[2]] == 100
        assert model.scaling_factor[model.v[3]] == 15
        assert model.scaling_factor[model.v[4]] == 20
        assert model.scaling_factor[model.c[1]] == 50
        assert model.scaling_factor[model.c[2]] == 250
        assert model.scaling_factor[model.c[3]] == 125
        assert model.scaling_factor[model.c[4]] == 625
        assert len(model.scaling_factor) == 8

        for bd in model.b.values():
            assert bd.scaling_factor[bd.v2] == 10
            assert len(bd.scaling_factor) == 1

        # Ensure we have not mutated original dict
        assert sdict == {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "unknown",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_block_data_overwrite_false(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "unknown",
        }

        scaling_factors_from_dict(model, sdict, overwrite=False, verify_names=True)

        assert model.scaling_factor[model.v[1]] == 5
        assert model.scaling_factor[model.v[2]] == 10
        assert model.scaling_factor[model.v[3]] == 15
        assert model.scaling_factor[model.v[4]] == 20
        assert model.scaling_factor[model.c[1]] == 5
        assert model.scaling_factor[model.c[2]] == 25
        assert model.scaling_factor[model.c[3]] == 125
        assert model.scaling_factor[model.c[4]] == 625
        assert len(model.scaling_factor) == 8

        for bd in model.b.values():
            assert bd.scaling_factor[bd.v2] == 10
            assert len(bd.scaling_factor) == 1

        # Ensure we have not mutated original dict
        assert sdict == {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "unknown",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_block_data_verify_fail(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "foo",
        }

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Block name (unknown) does not match that recorded in json_dict (foo)"
            ),
        ):
            scaling_factors_from_dict(model, sdict, overwrite=True, verify_names=True)

        # Ensure we have not mutated original dict
        assert sdict == {
            "v[1]": 50,
            "v[2]": 100,
            "c[1]": 50,
            "c[2]": 250,
            "block_name": "foo",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_indexed_block(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "block_datas": {
                "b[1]": {
                    "v2": 42,
                },
                "b[2]": {
                    "v2": 42,
                },
            },
            "block_name": "b",
        }

        scaling_factors_from_dict(model.b, sdict, overwrite=True, verify_names=True)

        assert model.scaling_factor[model.v[1]] == 5
        assert model.scaling_factor[model.v[2]] == 10
        assert model.scaling_factor[model.v[3]] == 15
        assert model.scaling_factor[model.v[4]] == 20
        assert model.scaling_factor[model.c[1]] == 5
        assert model.scaling_factor[model.c[2]] == 25
        assert model.scaling_factor[model.c[3]] == 125
        assert model.scaling_factor[model.c[4]] == 625
        assert len(model.scaling_factor) == 8

        for k in [1, 2]:
            assert model.b[k].scaling_factor[model.b[k].v2] == 42
            assert len(model.b[k].scaling_factor) == 1
        for k in [3, 4]:
            assert model.b[k].scaling_factor[model.b[k].v2] == 10
            assert len(model.b[k].scaling_factor) == 1

        # Ensure we have not mutated original dict
        assert sdict == {
            "block_datas": {
                "b[1]": {
                    "v2": 42,
                },
                "b[2]": {
                    "v2": 42,
                },
            },
            "block_name": "b",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_indexed_block_overwrite_false(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "block_datas": {
                "b[1]": {
                    "v2": 42,
                },
                "b[2]": {
                    "v2": 42,
                },
            },
            "block_name": "b",
        }

        scaling_factors_from_dict(model.b, sdict, overwrite=False, verify_names=True)

        assert model.scaling_factor[model.v[1]] == 5
        assert model.scaling_factor[model.v[2]] == 10
        assert model.scaling_factor[model.v[3]] == 15
        assert model.scaling_factor[model.v[4]] == 20
        assert model.scaling_factor[model.c[1]] == 5
        assert model.scaling_factor[model.c[2]] == 25
        assert model.scaling_factor[model.c[3]] == 125
        assert model.scaling_factor[model.c[4]] == 625
        assert len(model.scaling_factor) == 8

        for k in [1, 2, 3, 4]:
            assert model.b[k].scaling_factor[model.b[k].v2] == 10
            assert len(model.b[k].scaling_factor) == 1

        # Ensure we have not mutated original dict
        assert sdict == {
            "block_datas": {
                "b[1]": {
                    "v2": 42,
                },
                "b[2]": {
                    "v2": 42,
                },
            },
            "block_name": "b",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_verify_names_failure(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "block_datas": {
                "b[1]": {
                    "v2": 42,
                },
                "b[2]": {
                    "v2": 42,
                },
            },
            "block_name": "foo",
        }

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Block name (b) does not match that recorded in json_dict (foo)"
            ),
        ):
            scaling_factors_from_dict(model.b, sdict, overwrite=True, verify_names=True)

        # Ensure we have not mutated original dict
        assert sdict == {
            "block_datas": {
                "b[1]": {
                    "v2": 42,
                },
                "b[2]": {
                    "v2": 42,
                },
            },
            "block_name": "foo",
        }

    @pytest.mark.unit
    def test_scaling_factors_from_dict_invalid_component(self, model):
        # Partial dict of scaling factors to ensure we only touch things in the dict
        sdict = {
            "block_datas": {
                "b[1]": {
                    "v2": 42,
                },
                "b[2]": {
                    "v2": 42,
                },
            },
            "block_name": "foo",
        }

        with pytest.raises(
            TypeError, match=re.escape("v is not an instance of a Block of Suffix.")
        ):
            scaling_factors_from_dict(model.v, sdict, overwrite=True, verify_names=True)

    @pytest.mark.unit
    def test_scaling_factors_to_json_file(self, model):
        temp_context = TempfileManager.new_context()
        tmpfile = temp_context.create_tempfile(suffix=".json")

        scaling_factors_to_json_file(model, tmpfile)

        with open(tmpfile, "r") as f:
            lines = f.read()
        f.close()

        print(lines)

        expected = """{
   "v[1]": 5,
   "c[1]": 5,
   "v[2]": 10,
   "c[2]": 25,
   "v[3]": 15,
   "c[3]": 125,
   "v[4]": 20,
   "c[4]": 625,
   "subblock_suffixes": {
      "b[1]": {
         "v2": 10,
         "subblock_suffixes": {}
      },
      "b[2]": {
         "v2": 10,
         "subblock_suffixes": {}
      },
      "b[3]": {
         "v2": 10,
         "subblock_suffixes": {}
      },
      "b[4]": {
         "v2": 10,
         "subblock_suffixes": {}
      }
   },
   "block_name": "unknown"
}"""

        assert lines == expected

        # Check for clean up
        temp_context.release(remove=True)
        assert not os.path.exists(tmpfile)

    @pytest.mark.unit
    def test_scaling_factors_from_json_file(self, model):
        fname = os.path.join(currdir, "load_scaling_factors.json")

        scaling_factors_from_json_file(model, fname, overwrite=True)

        assert model.scaling_factor[model.v[1]] == 50
        assert model.scaling_factor[model.v[2]] == 100
        assert model.scaling_factor[model.v[3]] == 150
        assert model.scaling_factor[model.v[4]] == 200
        assert model.scaling_factor[model.c[1]] == 50
        assert model.scaling_factor[model.c[2]] == 250
        assert model.scaling_factor[model.c[3]] == 1250
        assert model.scaling_factor[model.c[4]] == 6250
        assert len(model.scaling_factor) == 8

        for k in [1, 2, 3, 4]:
            assert model.b[k].scaling_factor[model.b[k].v2] == 100
            assert len(model.b[k].scaling_factor) == 1


class TestGetScalingFactor:
    @pytest.mark.unit
    def test_get_scaling_factor_block(self):
        m = ConcreteModel()

        with pytest.raises(TypeError, match="Blocks cannot have scaling factors."):
            get_scaling_factor(m)

    @pytest.mark.unit
    def test_get_scaling_factor(self):
        m = ConcreteModel()
        m.v = Var()

        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        m.scaling_factor[m.v] = 10

        assert get_scaling_factor(m.v) == 10

    @pytest.mark.unit
    def test_get_scaling_factor_none(self):
        m = ConcreteModel()
        m.v = Var()

        m.scaling_factor = Suffix(direction=Suffix.EXPORT)

        assert get_scaling_factor(m.v) is None

    @pytest.mark.unit
    def test_get_scaling_factor_no_suffix(self):
        m = ConcreteModel()
        m.v = Var()

        assert get_scaling_factor(m.v) is None


class TestSetScalingFactor:
    @pytest.mark.unit
    def test_set_scaling_factor(self):
        m = ConcreteModel()
        m.scaling_factor = Suffix(direction=Suffix.EXPORT)

        m.v = Var()

        set_scaling_factor(m.v, 42)

        assert m.scaling_factor[m.v] == 42

    @pytest.mark.unit
    def test_set_scaling_factor_new_suffix(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v = Var()

        set_scaling_factor(m.v, 42)

        assert m.scaling_factor[m.v] == 42.0

        assert "Created new scaling suffix for unknown" in caplog.text

    @pytest.mark.unit
    def test_set_scaling_factor_not_float(self):
        m = ConcreteModel()
        m.v = Var()

        with pytest.raises(
            ValueError, match="could not convert string to float: 'foo'"
        ):
            set_scaling_factor(m.v, "foo")

    @pytest.mark.unit
    def test_set_scaling_factor_negative(self):
        m = ConcreteModel()
        m.v = Var()

        with pytest.raises(
            ValueError,
            match=re.escape(
                "scaling factor for v is negative (-42.0). "
                "Scaling factors must be strictly positive."
            ),
        ):
            set_scaling_factor(m.v, -42)

    @pytest.mark.unit
    def test_set_scaling_factor_zero(self):
        m = ConcreteModel()
        m.v = Var()

        with pytest.raises(
            ValueError,
            match=re.escape(
                "scaling factor for v is zero. "
                "Scaling factors must be strictly positive."
            ),
        ):
            set_scaling_factor(m.v, 0)

    @pytest.mark.unit
    def test_set_scaling_factor_overwrite_false(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v = Var()

        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        m.scaling_factor[m.v] = 10

        set_scaling_factor(m.v, 42, overwrite=False)

        assert (
            "Existing scaling factor for v found and overwrite=False. "
            "Scaling factor unchanged." in caplog.text
        )
        assert m.scaling_factor[m.v] == 10


class TestDelScalingFactor:
    @pytest.mark.unit
    def test_del_scaling_factor(self):
        m = ConcreteModel()
        m.v = Var()

        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        m.scaling_factor[m.v] = 10

        del_scaling_factor(m.v)

        assert len(m.scaling_factor) == 0

    @pytest.mark.unit
    def test_del_scaling_factor_not_present(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v = Var()

        m.scaling_factor = Suffix(direction=Suffix.EXPORT)

        del_scaling_factor(m.v)

        assert len(m.scaling_factor) == 0

    @pytest.mark.unit
    def test_del_scaling_factor_delete_empty(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v = Var()

        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        m.scaling_factor[m.v] = 10

        del_scaling_factor(m.v, delete_empty_suffix=True)

        assert not hasattr(m, "scaling_factor")

        assert "Deleting empty scaling suffix from unknown" in caplog.text


class TestReportScalingFactors:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.s = Set(initialize=[1, 2, 3, 4])

        m.v = Var(m.s)

        @m.Constraint(m.s)
        def c(b, i):
            return b.v[i] == i

        m.b = Block(m.s)

        for i, bd in m.b.items():
            bd.v2 = Var()

        # Need to check all possible behaviours
        # Set values for half the variables (indexes 1 and 3)
        for i in [1, 3]:
            m.v[i].set_value(42)
            m.b[i].v2.set_value(42)

        # Set scaling factors for half the components (indexed 1 and 2)
        m.scaling_factor = Suffix(direction=Suffix.EXPORT)
        for i in [1, 2]:
            m.scaling_factor[m.v[i]] = 5 * i
            m.scaling_factor[m.c[i]] = 5**i
            m.b[i].scaling_factor = Suffix(direction=Suffix.EXPORT)
            m.b[i].scaling_factor[m.b[i].v2] = 10

        return m

    @pytest.mark.unit
    def test_report_scaling_factors_all(self, model):
        stream = StringIO()

        report_scaling_factors(model, descend_into=True, stream=stream)

        expected = """Scaling Factors for unknown

Variable    Scaling Factor    Value        Scaled Value
v[1]        5.000E+00         4.200E+01    2.100E+02
v[2]        1.000E+01         None         None
v[3]        None              4.200E+01    4.200E+01
v[4]        None              None         None
b[1].v2     1.000E+01         4.200E+01    4.200E+02
b[2].v2     1.000E+01         None         None
b[3].v2     None              4.200E+01    4.200E+01
b[4].v2     None              None         None

Constraint    Scaling Factor
c[1]          5.000E+00
c[2]          2.500E+01
c[3]          None
c[4]          None
"""

        print(stream.getvalue())
        assert stream.getvalue() == expected

    @pytest.mark.unit
    def test_report_scaling_factors_descend_false(self, model):
        stream = StringIO()

        report_scaling_factors(model, descend_into=False, stream=stream)

        expected = """Scaling Factors for unknown

Variable    Scaling Factor    Value        Scaled Value
v[1]        5.000E+00         4.200E+01    2.100E+02
v[2]        1.000E+01         None         None
v[3]        None              4.200E+01    4.200E+01
v[4]        None              None         None

Constraint    Scaling Factor
c[1]          5.000E+00
c[2]          2.500E+01
c[3]          None
c[4]          None
"""

        assert stream.getvalue() == expected

    @pytest.mark.unit
    def test_report_scaling_factors_vars_only(self, model):
        stream = StringIO()

        report_scaling_factors(model, descend_into=True, ctype=Var, stream=stream)

        expected = """Scaling Factors for unknown

Variable    Scaling Factor    Value        Scaled Value
v[1]        5.000E+00         4.200E+01    2.100E+02
v[2]        1.000E+01         None         None
v[3]        None              4.200E+01    4.200E+01
v[4]        None              None         None
b[1].v2     1.000E+01         4.200E+01    4.200E+02
b[2].v2     1.000E+01         None         None
b[3].v2     None              4.200E+01    4.200E+01
b[4].v2     None              None         None
"""

        assert stream.getvalue() == expected

    @pytest.mark.unit
    def test_report_scaling_factors_constraints_only(self, model):
        stream = StringIO()

        report_scaling_factors(
            model, descend_into=True, stream=stream, ctype=Constraint
        )

        expected = """Scaling Factors for unknown

Constraint    Scaling Factor
c[1]          5.000E+00
c[2]          2.500E+01
c[3]          None
c[4]          None
"""

        assert stream.getvalue() == expected

    @pytest.mark.unit
    def test_report_scaling_factors_indexed_block(self, model):
        stream = StringIO()

        report_scaling_factors(model.b, descend_into=True, stream=stream)

        expected = """Scaling Factors for b

Variable    Scaling Factor    Value        Scaled Value
b[1].v2     1.000E+01         4.200E+01    4.200E+02
b[2].v2     1.000E+01         None         None
b[3].v2     None              4.200E+01    4.200E+01
b[4].v2     None              None         None
"""

        assert stream.getvalue() == expected

    @pytest.mark.unit
    def test_report_scaling_factors_not_block(self, model):
        stream = StringIO()

        with pytest.raises(
            TypeError,
            match="report_scaling_factors: blk must be an instance of a Pyomo Block.",
        ):
            report_scaling_factors(
                model.v, descend_into=True, stream=stream, ctype=Constraint
            )

    @pytest.mark.unit
    def test_report_scaling_factors_invalid_ctype(self, model):
        stream = StringIO()

        with pytest.raises(
            ValueError,
            match="report_scaling_factors only supports None, Var or Constraint for argument ctype: "
            "received foo.",
        ):
            report_scaling_factors(model, descend_into=True, stream=stream, ctype="foo")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Pure component gas viscosities at low pressures from critical properties,
acentric factor, and dipole moment. Also requires an association factor, which
can be found for some highly polar substances in Table 9-1 in the Properties
of Gases and Liquids, 5th Ed. Chung et al. might also have additional factors
in some of their papers. If unknown, set the association factor to zero.
"""
from pyomo.environ import Var, units as pyunits
import pyomo.environ as pyo

from idaes.core.util.misc import set_param_from_config
from idaes.models.properties.modular_properties.pure.ChapmanEnskog import (
    collision_integral_neufeld_callback,
)


class ChungViscosityPure(object):
    """Container class for physical property calculations via methods of Chung et al."""

    @staticmethod
    def build_common_parameters(cobj):
        """Build dipole moment and association factor"""
        units = cobj.parent_block().get_metadata().derived_units
        if not hasattr(cobj, "dipole_moment"):
            cobj.dipole_moment = Var(
                doc="Molecular dipole moment",
                units=units["current"] * units["time"] * units["length"],
            )
            set_param_from_config(cobj, param="dipole_moment")

        if not hasattr(cobj, "association_factor_chung"):
            cobj.association_factor_chung = Var(
                doc="Correction term used for highly polar components",
                units=pyunits.dimensionless,
            )
            set_param_from_config(cobj, param="association_factor_chung")

    class visc_d_phase_comp(object):
        """
        Implementation of pure component gas dynamic viscosity via the method of Chung et al. as described in The Properties
         of Gases and Liquids, Section 9-4-2.
        """

        @staticmethod
        def build_parameters(cobj, p):
            """Build common parameters and viscosity collision integral callback"""
            ChungViscosityPure.build_common_parameters(cobj)
            if not hasattr(cobj, "viscosity_collision_integral_callback"):
                cobj.viscosity_collision_integral_callback = (
                    collision_integral_neufeld_callback
                )

        @staticmethod
        def return_expression(b, cobj, p, T):
            """Return expression for visc_d_phase_comp"""
            # Properties of Gases and Liquids 5th Ed., Section 9-4-2
            units = b.params.get_metadata().derived_units

            T = pyunits.convert(T, to_units=pyunits.K)
            T_crit = pyunits.convert(cobj.temperature_crit, to_units=pyunits.K)
            V_crit = 1 / pyunits.convert(
                cobj.dens_mol_crit, to_units=pyunits.mol / pyunits.mL
            )
            M = pyunits.convert(cobj.mw, pyunits.g / pyunits.mol)
            mu = pyunits.convert(cobj.dipole_moment, pyunits.debye)
            omega = cobj.omega
            kappa = cobj.association_factor_chung
            Omega = cobj.viscosity_collision_integral_callback(
                1.2593 * T / T_crit
            )  # Eq. 9-4.10
            # Eq. 9-4.12
            mu_r = (
                131.3
                * pyo.sqrt(pyunits.mL / pyunits.mol * pyunits.K)
                / pyunits.debye
                * (mu / pyo.sqrt(V_crit * T_crit))
            )
            # Eq. 9-4.11
            Fc = 1 - 0.2756 * omega + 0.059035 * mu_r**4 + kappa
            C = (
                40.785
                * pyunits.micropoise
                * (pyunits.mL / pyunits.mol) ** (2 / 3)
                / pyo.sqrt(pyunits.g / pyunits.mol * pyunits.K)
            )
            # Eq. 9-4.10
            visc = C * Fc * pyo.sqrt(M * T) / (V_crit ** (2 / 3) * Omega)

            return pyunits.convert(visc, units["dynamic_viscosity"])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Authors: Anuja Deshpande, Andrew Lee
"""
import pytest
import numpy as np

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Set,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.core import Component

from idaes.core import FlowsheetBlock

from idaes.models.unit_models import Flash

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE

from idaes.models.properties.modular_properties.examples.CO2_H2O_Ideal_VLE import (
    configuration,
)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


def _as_quantity(x):
    unit = pyunits.get_units(x)
    if unit is None:
        unit = pyunits.dimensionless
    return value(x) * unit._get_pint_unit()


class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 2
        for i in model.params.component_list:
            assert i in ["H2O", "CO2"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 3
        for i in model.params._phase_component_set:
            assert i in [("Liq", "H2O"), ("Vap", "H2O"), ("Vap", "CO2")]

        assert model.params.config.state_definition == FTPx

        assertStructuredAlmostEqual(
            model.params.config.state_bounds,
            {
                "flow_mol": (0, 10, 20, pyunits.mol / pyunits.s),
                "temperature": (273.15, 323.15, 1000, pyunits.K),
                "pressure": (5e4, 108900, 1e7, pyunits.Pa),
            },
            item_callback=_as_quantity,
        )

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE
        }

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 1
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1"]

        assert model.params.phase_equilibrium_list == {"PE1": ["H2O", ("Vap", "Liq")]}

        assert model.params.pressure_ref.value == 101325
        assert model.params.temperature_ref.value == 298.15

        assert model.params.H2O.mw.value == 18.0153e-3
        assert model.params.H2O.pressure_crit.value == 220.64e5
        assert model.params.H2O.temperature_crit.value == 647

        assert model.params.CO2.mw.value == 44.0095e-3
        assert model.params.CO2.pressure_crit.value == 73.825e5
        assert model.params.CO2.temperature_crit.value == 304.23

        assert_units_consistent(model)


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        model.props = model.params.build_state_block([1], defined_state=True)

        model.props[1].flow_mol.fix(10)
        model.props[1].temperature.fix(323.15)
        model.props[1].pressure.fix(108900)
        model.props[1].mole_frac_comp["H2O"].fix(0.5)
        model.props[1].mole_frac_comp["CO2"].fix(0.5)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        # Check state variable values and bounds
        assert isinstance(model.props[1].flow_mol, Var)
        assert value(model.props[1].flow_mol) == 10
        assert model.props[1].flow_mol.ub == 20
        assert model.props[1].flow_mol.lb == 0

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 108900
        assert model.props[1].pressure.ub == 1e7
        assert model.props[1].pressure.lb == 5e4

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 323.15
        assert model.props[1].temperature.ub == 1000
        assert model.props[1].temperature.lb == 273.15

        assert isinstance(model.props[1].mole_frac_comp, Var)
        assert len(model.props[1].mole_frac_comp) == 2
        assert value(model.props[1].mole_frac_comp["H2O"]) == 0.5
        assert value(model.props[1].mole_frac_comp["CO2"]) == 0.5

        assert_units_consistent(model)

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in [
                "Total Molar Flowrate",
                "Total Mole Fraction",
                "Temperature",
                "Pressure",
            ]

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model.props[1]) == 0

    @pytest.mark.component
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.props[1].mole_frac_phase_comp["Liq", "H2O"].value == pytest.approx(
            1, abs=1e-4
        )
        assert model.props[1].mole_frac_phase_comp["Vap", "H2O"].value == pytest.approx(
            0.11501, abs=1e-4
        )
        assert model.props[1].mole_frac_phase_comp["Vap", "CO2"].value == pytest.approx(
            0.88499, abs=1e-4
        )
        assert model.props[1].phase_frac["Vap"].value == pytest.approx(
            0.56498, abs=1e-4
        )

    @pytest.mark.integration
    def test_temp_swing(self):
        # Create a flash model with the CO2-H2O property package
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = GenericParameterBlock(**configuration)
        m.fs.flash = Flash(property_package=m.fs.properties)

        # Fix inlet stream state variables
        m.fs.flash.inlet.flow_mol.fix(9.89433124673833)  # mol/s
        m.fs.flash.inlet.mole_frac_comp[0, "CO2"].fix(0.13805801934749645)
        m.fs.flash.inlet.mole_frac_comp[0, "H2O"].fix(0.8619419806525035)
        m.fs.flash.inlet.pressure.fix(183430)  # Pa
        m.fs.flash.inlet.temperature.fix(396.79057912844183)  # K

        # Fix flash and its outlet conditions
        m.fs.flash.deltaP.fix(0)
        m.fs.flash.vap_outlet.temperature.fix(313.15)

        # Initialize the flash model
        m.fs.flash.initialize()

        # Create a dictionary of expected solution for flash outlet temperature
        # sweep
        temp_range = list(np.linspace(313, 396))

        expected_vapor_frac = [
            0.14388,
            0.14445,
            0.14508,
            0.14576,
            0.1465,
            0.14731,
            0.14818,
            0.14913,
            0.15017,
            0.15129,
            0.15251,
            0.15384,
            0.15528,
            0.15685,
            0.15856,
            0.16042,
            0.16245,
            0.16467,
            0.16709,
            0.16974,
            0.17265,
            0.17584,
            0.17935,
            0.18323,
            0.18751,
            0.19226,
            0.19755,
            0.20346,
            0.21008,
            0.21755,
            0.22601,
            0.23565,
            0.24673,
            0.25956,
            0.27456,
            0.29229,
            0.31354,
            0.33942,
            0.37157,
            0.4125,
            0.46628,
            0.53993,
            0.64678,
            0.81547,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
        ]

        expected_heat_duty = [
            -396276.55508,
            -394859.42896,
            -393421.28226,
            -391960.40602,
            -390474.94512,
            -388962.88119,
            -387422.01297,
            -385849.93371,
            -384244.00483,
            -382601.32549,
            -380918.69691,
            -379192.58068,
            -377419.04976,
            -375593.73054,
            -373711.73419,
            -371767.57480,
            -369755.07135,
            -367667.22968,
            -365496.09940,
            -363232.59958,
            -360866.30485,
            -358385.18097,
            -355775.25563,
            -353020.20490,
            -350100.82935,
            -346994.38367,
            -343673.70980,
            -340106.10300,
            -336251.80960,
            -332062.00898,
            -327476.06061,
            -322417.68382,
            -316789.55435,
            -310465.49473,
            -303278.90949,
            -295005.17406,
            -285333.93764,
            -273823.89005,
            -259825.51107,
            -242341.82570,
            -219760.16114,
            -189290.02362,
            -145647.55666,
            -77469.59283,
            -3219.88910,
            -2631.66067,
            -2043.09220,
            -1454.17760,
            -864.91585,
            -275.30623,
        ]

        outvals = zip(expected_vapor_frac, expected_heat_duty)
        expected_sol = dict(zip(temp_range, outvals))

        # Solve the model for a range of flash outlet temperatures
        # Perform flash outlet temperature sweep and test the solution
        for t in temp_range:
            m.fs.flash.vap_outlet.temperature.fix(t)
            res = solver.solve(m)
            assert res.solver.termination_condition == "optimal"
            frac = value(m.fs.flash.vap_outlet.flow_mol[0]) / value(
                m.fs.flash.inlet.flow_mol[0]
            )
            assert frac == pytest.approx(expected_sol[t][0], abs=1e-4)
            hduty = value(m.fs.flash.heat_duty[0])
            assert hduty == pytest.approx(expected_sol[t][1], rel=1e-4)

    @pytest.mark.unit
    def test_report(self, model):
        model.props[1].report()
